import os,sys,pickle,re,importlib
import subprocess as sp, pandas as pd,numpy as np,time, pickle, os
import dorianUtils.utilsD as ut
import plotly.express as px
import plotly.graph_objects as go
from dorianUtils.utilsD import Utils
utils=Utils()
from dorianUtils.comUtils import (html_table,timenowstd,print_file,computetimeshow)
import dorianUtils.VersionsManager as vm
from scipy import signal
importlib.reload(vm)

class VersionsManager_extend(vm.VersionsManager_daily):
    def __init__(self,*args,**kwargs):
        vm.VersionsManager_daily.__init__(self,*args,**kwargs)
        _v_start={
            1.0:'2022-07-04 15:00',
            2.0:'2022-07-06 7:00',
            }
        self.versions = pd.concat([self.versions,pd.Series(_v_start,name='start')],axis=1)


file_transition='/home/dorian/data/sylfenData/test_plc_folder/versionnage_tags.ods'
dir_plc='/home/dorian/data/sylfenData/test_plc_folder/'
folderPkl='/home/dorian/data/sylfenData/test_daily_folder/'

svm = VersionsManager_extend(folderPkl,dir_plc,file_transition,pattern_plcFiles='*plc*.xlsx')

# svm.generate_versionning_data(False,True)
svm.generate_versionning_data(True,True)

day='2022-07-06'
######################
# GENERATE THE DATA  #
######################
# def generate_the_data():
data={}
ts=pd.date_range(day+' 8:00',day+' 22:00',freq='30s',tz='CET')
s_empty=pd.Series([],name='value')
phi=np.linspace(0,10,len(ts))
data['tag_FT_01']=pd.Series(signal.square(phi),index=ts,name='value')
data['tag_01_FT']=pd.Series(0.5,index=ts,name='value')
data['tag_02_FT']=pd.Series(signal.sawtooth(phi),index=ts,name='value')
data['tag_FT_03']=s_empty.copy()
data['tag_03_FT']=pd.Series(0.7,index=ts,name='value')
data['tag_FT_04']=pd.Series(np.sin(phi),index=ts,name='value')
data['tag_04_FT']=s_empty.copy()
data['tag1']=pd.Series(phi,index=ts,name='value')/10
data['tag3']=s_empty.copy()
data['tag4']=pd.Series(0.2,index=ts,name='value')
data['tag7']=pd.Series(np.log(0.5+phi),index=ts,name='value')/2

for k,s in data.items():s.to_pickle(folderPkl+'/'+day+'/'+k+'.pkl')
###################### #
# look at THE DATA     #
# and see the presence #
#       of tags        #
###################### #
def show_data(day):
    data2={t.strip('.pkl'):pd.read_pickle(folderPkl+'/'+day+'/'+t) for t in os.listdir(folderPkl+'/'+day)}
    data2={k:v for k,v in data2.items() if not v.empty}
    df=pd.concat(data2,axis=1)
    df=df[df.columns.sort_values()]
    fig=px.line(df)
    colors={}
    colors['tag_01_FT']= 'red'
    colors['tag_FT_01']= 'pink'
    colors['tag_02_FT']= 'purple'
    colors['tag_FT_02']= 'magenta'
    colors['tag_FT_03']= 'blue'
    colors['tag_03_FT']= 'cyan'
    colors['tag_FT_04']= 'orange'
    colors['tag_04_FT']= 'yellow'
    colors['tag1']= 'black'
    colors['tag2']= 'black'
    colors['tag3']= 'black'
    colors['tag4']= 'green'
    colors['tag5']= 'black'
    colors['tag6']= 'black'
    colors['tag7']= 'grey'

    for k in fig.data:
        if k.name in colors:
            k.marker.color=colors[k.name]
            k.line.color=colors[k.name]
    fig.show()

show_data('2022-07-06')
# svm.presence_tags(data.keys(),empty_df=True).show()


#################
# DIAGNOSTIQUER #
# LA SITUATION  #
#################
# df_missing_tags=svm._load_missing_tags_map()
# svm.show_map_of_compatibility(df_missing_tags)

####################
# OBTENIR LA TABLE #
# DE  RENNOMMAGE   #
####################
transition='1_2'
# svm.get_rename_tags_map_from_rules(transition)
# svm.make_period_compatible_from_transition(day,day,transition,False)
# v=1.0;html_table(svm.df_plcs[v].loc[svm.missing_tags_versions('2022-07-06',v)[v]])
# v=2.0;html_table(svm.df_plcs[v].loc[svm.missing_tags_versions('2022-07-06',v)[v]])

####################
# RENDRE COMPATIBE #
####################
svm.create_missing_tags_period_version('2022-07-05','2022-07-10',1.0)
svm.make_period_compatible_from_transition('2022-07-05','2022-07-10',transition,True)
show_data('2022-07-06')

sys.exit()



import pandas as pd,numpy as np
from multiprocessing import Process, Queue, current_process,Pool
from pandas.tseries.frequencies import to_offset
import subprocess as sp, os,re, pickle,glob
import time, datetime as dt, pytz
from scipy import linalg, integrate
from dateutil import parser
from dorianUtils.utilsD import Utils
import plotly.express as px

pd.options.mode.chained_assignment = None  # default='warn'

class ConfigMaster:
    """docstring for ConfigMaster."""

    def __init__(self,folderPkl):
        self.utils      = Utils()
        self.folderPkl  = folderPkl
        self.listFiles  = self.utils.get_listFilesPklV2(folderPkl)
        # self.dfPLC = self._loadDF_PLC()
# ==============================================================================
#                                 functions
# ==============================================================================
    def getUnitofTag(self,tag):
        unit=self.dfPLC.loc[tag].UNITE
        # print(unit)
        if not isinstance(unit,str):
            unit='u.a'
        return unit

    def getTagsTU(self,patTag,units=None,onCol='index',cols='tag'):
        #patTag
        if onCol=='index':
            df = self.dfPLC[self.dfPLC.index.str.contains(patTag,case=False)]
        else:
            df = self.dfPLC[self.dfPLC[onCol].str.contains(patTag,case=False)]

        #units
        if not units : units = self.listUnits
        if isinstance(units,str):units = [units]
        df = df[df['UNITE'].isin(units)]

        #return
        if cols=='tdu' :
            return df[['DESCRIPTION','UNITE']]
        elif cols=='tag':
            return list(df.index)
        else :
            return df

    def loadFile(self,filename,skip=1):
        if '*' in filename :
            filenames=self.utils.get_listFilesPklV2(self.folderPkl,filename)
            if len(filenames)>0 : filename=filenames[0]
            else : return pd.DataFrame()
        df = pickle.load(open(filename, "rb" ))
        return df[::skip]

    def _parkTag(self,df,tag,folder):
        print(tag)
        # colTag = [k for k in df.columns if k.lower()=='tag']
        dfTag=df[df.tag==tag]
        with open(folder + tag + '.pkl' , 'wb') as handle:
            pickle.dump(dfTag, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def parkDayPKL(self,datum,offset=1,pool=False):
        print(datum)
        realDatum=parser.parse(datum)+dt.timedelta(days=offset)
        df = self.loadFile('*'+ realDatum.strftime('%Y-%m-%d') + '*')
        if not df.empty:
            folder=self.folderPkl+'parkedData/'+ datum + '/'
            if not os.path.exists(folder):os.mkdir(folder)
            # listTags=list(self.dfPLC.TAG.unique())
            listTags=list(df.tag.unique())
            if pool:
                with Pool() as p:p.starmap(self._parkTag,[(df,tag,folder) for tag in listTags])
            else:
                for tag in listTags:
                    self._parkTag(df,tag,folder)
        else : print('no data loaded')

    def parkDates(self,listDates,nCores=4,offset=0):
        if nCores>1:
            with Pool(nCores) as p:
                p.starmap(self.parkDayPKL,[(datum,offset,False) for datum in listDates])
        else :
            for datum in listDates:
                self.parkDayPKL(datum)

    def updateLayoutStandard(self,fig,sizeDots=5):
        fig.update_yaxes(showgrid=False)
        fig.update_traces(selector=dict(type='scatter'),marker=dict(size=sizeDots))
        fig.update_layout(height=750)
        # fig.update_traces(hovertemplate='<b>%{y:.2f}')
        fig.update_traces(hovertemplate='     <b>%{y:.2f}<br>     %{x|%H:%M:%S}')
        return fig

    def update_lineshape(self,fig,style='default'):
        if style in ['markers','lines','lines+markers']:
            fig.update_traces(line_shape="linear",mode=style)
        elif style =='stairs':
            fig.update_traces(line_shape="hv",mode='lines')
        return fig

class ConfigDashTagUnitTimestamp(ConfigMaster):
    def __init__(self,folderPkl,confFolder,encode='utf-8'):
        ConfigMaster.__init__(self,folderPkl)
        self.confFolder   = confFolder
        self.confFile     = glob.glob(self.confFolder + '*PLC*')[0]
        self.modelAndFile = self.__getModelNumber()
        self.listFilesPkl = self._get_ValidFiles()
        self.parked   = False
        if os.path.exists(self.folderPkl+'/parkedData'):
            self.parked     = True
            self.parkedDays =[k.split('/')[-1] for k in glob.glob(self.folderPkl+'parkedData/*')]
            self.parkedDays.sort()
        self.dfPLC        = pd.read_csv(self.confFile,encoding=encode)
        try :
            self.usefulTags = pd.read_csv(self.confFolder+'/predefinedCategories.csv',index_col=0)
            self.usefulTags.index = self.utils.uniformListStrings(list(self.usefulTags.index))
        except :
            self.usefulTags = pd.DataFrame()

        self.unitCol,self.descriptCol,self.tagCol = self._getPLC_ColName()
        self.listUnits    = self._get_UnitsdfPLC()

    def __getModelNumber(self):
        modelNb = re.findall('\d{5}-\d{3}',self.confFile)
        if not modelNb: return ''
        else : return modelNb[0]

# ==============================================================================
#                     basic functions
# ==============================================================================
    def _getPLC_ColName(self):
        l = self.dfPLC.columns
        v = l.to_frame()
        unitCol = ['unit' in k.lower() for k in l]
        descriptName = ['descript' in k.lower() for k in l]
        tagName = ['tag' in k.lower() for k in l]
        return [list(v[k][0])[0] for k in [unitCol,descriptName,tagName]]

    def _get_ValidFiles(self):
        return self.utils.get_listFilesPklV2(self.folderPkl)

    def _get_UnitsdfPLC(self):
        listUnits = self.dfPLC[self.unitCol]
        return listUnits[~listUnits.isna()].unique().tolist()

    def getUsefulTags(self,usefulTag):
        category = self.usefulTags.loc[usefulTag]
        return self.getTagsTU(category.Pattern,category.Unit)

    def checkDiffbetweenPLCandDF(self,df):
        '''difference between tags in df and dfPLC'''
        listTagsDf = pd.DataFrame(df.tag.unique(),columns=['tag'])
        listTagsDf = listTagsDf.sort_values(by='tag')
        dfplc = pd.DataFrame(self.dfPLC.TAG.sort_values())
        dfplc['id']='dfplc'
        listTagsDf['id']='listTagsDf'
        listTagsDf.columns=['TAG','id']
        return pd.concat([dfplc,listTagsDf],axis=0).drop_duplicates(subset='TAG',keep=False)

    def checkDFvsPLCTags(self,df):
        df2=df.tag.unique()
        df2.columns='TAG'
        dupl = pd.concat([df2,self.dfPLC.TAG]).duplicated(subset=['TAG'])
        return dupl

    def renameColsPkl(self):
        import pickle
        for f in self.listFilesPkl:
            print(f)
            df=self.loadFile(f)
            df.columns=['tag','value','timestampUTC']
            with open(f, 'wb') as handle:
                pickle.dump(df, handle, protocol=pickle.HIGHEST_PROTOCOL)

# ==============================================================================
#                   functions filter on configuration file with tags
# ==============================================================================
    def DF_standardProcessing(self,df,rs='auto',applyMethod='mean',timezone='Europe/Paris'):
        if df.duplicated().any():
            df = df.drop_duplicates()
            print("attention il y a des doublons dans la base de donnees Jules")
        df['timestampz'] = df.timestampz.dt.tz_convert(timezone)
        df = df.pivot(index="timestampz", columns="tag", values="value")
        deltaT=(df.index[-1]-df.index[0]).total_seconds()
        if rs=='auto':rs = '{:.0f}'.format(max(1,deltaT//1000)) + 's'
        df['value'] = pd.to_numeric(df['value'],errors='coerce')
        df = df.sort_values(by=['timestampz','tag'])
        df = eval("df.resample('100ms').ffill().ffill().resample(rs).apply(np." + applyMethod + ")")
        return df

    def getUnitsOfpivotedDF(self,df,asList=True):
        dfP,listTags,tagsWithUnits=self.dfPLC,df.columns,[]
        for tagName in listTags :
            tagsWithUnits.append(dfP[dfP[self.tagCol]==tagName][[self.tagCol,self.unitCol]])
        tagsWithUnits = pd.concat(tagsWithUnits)
        if asList :tagsWithUnits = [k + ' ( in ' + l + ')' for k,l in zip(tagsWithUnits[self.tagCol],tagsWithUnits[self.unitCol])]
        return tagsWithUnits

    def getCatsFromUnit(self,unitName,pattern=None):
        if not pattern:pattern = self.listPatterns[0]
        sameUnitTags = self.getTagsTU('',unitName)
        return list(pd.DataFrame([re.findall(pattern,k)[0] for k in sameUnitTags])[0].unique())

    def getDescriptionFromTagname(self,tagName):
        return list(self.dfPLC[self.dfPLC[self.tagCol]==tagName][self.descriptCol])[0]

    def getTagnamefromDescription(self,desName):
        return list(self.dfPLC[self.dfPLC[self.descriptCol]==desName][self.tagCol])[0]

    def get_randomListTags(self,n=5):
        import random
        allTags = self.getTagsTU('',ds=True)
        return [allTags[random.randint(0,len(allTags))] for k in range(n)]


    # ==============================================================================
    #                   functions filter on dataFrame
    # ==============================================================================
    def _DF_fromTagList(self,datum,tagList,rs):
        realDatum = self.utils.datesBetween2Dates([datum,datum],offset=+1)[0]
        df=self.loadFile('*' + realDatum[0]  + '*')
        df = df.drop_duplicates(subset=['timestampUTC', 'tag'], keep='last')

        if not isinstance(tagList,list):tagList =[tagList]
        df = df[df.tag.isin(tagList)]
        if not rs=='raw':df = df.pivot(index="timestampUTC", columns="tag", values="value")
        else : df = df.sort_values(by=['tag','timestampUTC']).set_index('timestampUTC')
        return df

    def _DF_cutTimeRange(self,df,timeRange,timezone='Europe/Paris'):
        t0 = parser.parse(timeRange[0]).astimezone(pytz.timezone('UTC'))
        t1 = parser.parse(timeRange[1]).astimezone(pytz.timezone('UTC'))
        df=df[(df.index>t0)&(df.index<t1)].sort_index()
        df.index=df.index.tz_convert(timezone)# convert utc to tzSel timezone
        return df

    def _loadDFTagDay(self,datum,tag,raw=False):
        folderDaySmallPower=self.folderPkl+'parkedData/'+ datum + '/'
        try:
            df = pickle.load(open(folderDaySmallPower + tag + '.pkl', "rb" ))
            df = df.drop_duplicates(subset=['timestampUTC', 'tag'], keep='last')
            # df.duplicated(subset=['timestampUTC', 'tag'], keep=False)
            if not raw:
                # df = df.pivot(index="timestampUTC", columns="tag", values="value")
                df = df[['timestampUTC','value']].set_index('timestampUTC')
                df.columns=[tag]
            else : df=df.set_index('timestampUTC')
        except :
            df = pd.DataFrame()
            print('loading of file :',folderDaySmallPower + tag + '.pkl' ,' was not possible')
        return df

    def _loadDFTagsDay(self,datum,listTags,parked,pool,raw=False):
        dfs=[]
        print(datum)
        if parked :
            if pool :
                print('pooled process')
                with Pool() as p:
                    dfs=p.starmap(self._loadDFTagDay, [(datum,tag,raw) for tag in listTags])
            else :
                for tag in listTags:
                    dfs.append(self._loadDFTagDay(datum,tag,raw))
        else :
            dfs.append(self._DF_fromTagList(datum,listTags,raw))
        if raw:df  = pd.concat(dfs,axis=0)
        else :
            df = pd.concat(dfs,axis=1)
        return df

    def DF_loadTimeRangeTags(self,timeRange,listTags,rs='auto',applyMethod='mean',
                                parked=True,timezone='Europe/Paris',pool=True):
        if not timeRange:
            if parked : day = self.parkedDays[-1]
            else : day = re.findall('\d{4}-\d{2}-\d{2}',self.listFilesPkl[-1])[0]
            timeRange = [day + ' 9:00',day + ' 18:00']

        listDates,delta = self.utils.datesBetween2Dates(timeRange,offset=0)

        if rs=='auto':rs = '{:.0f}'.format(max(1,delta.total_seconds()//1000)) + 's'
        dfs=[]
        if len(listTags)>0:
            if pool:
                with Pool() as p:
                    dfs=p.starmap(self._loadDFTagsDay, [(datum,listTags,parked,False,rs=='raw') for datum in listDates])
            else:
                for datum in listDates:
                    dfs.append(self._loadDFTagsDay(datum,listTags,parked,True,rs=='raw'))
        else:
            return pd.DataFrame()
        if len(dfs)>0:
            df = pd.concat(dfs,axis=0)
            print("finish loading")
            if not df.empty:
                if not rs=='raw':
                    df = self._DF_cutTimeRange(df,timeRange,timezone)
                    df = df[~df.index.duplicated()]
                    df = eval("df.resample('100ms').ffill().ffill().resample(rs).apply(np." + applyMethod + ")")
                    df = df.loc[:,~df.columns.duplicated()]
                    df = df.reindex(sorted(df.columns),axis=1)
                else:
                    df['timestamp']=df.index
                    df = df.sort_values(by=['tag','timestamp'])
                    df = df.drop(['timestamp'],axis=1)
                    df = self._DF_cutTimeRange(df,timeRange,timezone)
            else : df= pd.DataFrame()
        else : df= pd.DataFrame()
        return df

    def plotMultiUnitGraph(self,timeRange,listTags=[],**kwargs):
        if not listTags : listTags=self.get_randomListTags()
        tagMapping = {t:self.getUnitofTag(t) for t in listTags}
        df  = self.DF_loadTimeRangeTags(timeRange,listTags,**kwargs)
        return self.utils.multiUnitGraph(df,tagMapping)

    def plotTabSelectedData(self,df):
        fig = px.scatter(df)
        unit = self.getUnitofTag(df.columns[0])
        nameGrandeur = self.utils.detectUnit(unit)
        return fig.update_layout(yaxis_title = nameGrandeur + ' in ' + unit)

    # ==============================================================================
    #                   functions to compute new variables
    # ==============================================================================

    def integrateDFTag(self,df,tagname,timeWindow=60,formatted=1):
        dfres = df[df.Tag==tagname]
        if not formatted :
            dfres = self.formatRawDF(dfres)
        dfres = dfres.sort_values(by='timestamp')
        dfres.index=dfres.timestamp
        dfres=dfres.resample('100ms').ffill()
        dfres=dfres.resample(str(timeWindow) + 's').mean()
        dfres['Tag'] = tagname
        return dfres

    def integrateDF(self,df,pattern,**kwargs):
        ts = time.time()
        dfs = []
        listTags = self.getTagsTU(pattern[0],pattern[1])[self.tagCol]
        # print(listTags)
        for tag in listTags:
            dfs.append(self.integrateDFTag(df,tagname=tag,**kwargs))
        print('integration of pattern : ',pattern[0],' finished in ')
        self.utils.printCTime(ts)
        return pd.concat(dfs,axis=0)

    # ==============================================================================
    #                   functions computation
    # ==============================================================================

    def fitDataframe(self,df,func='expDown',plotYes=True,**kwargs):
        res = {}
        period = re.findall('\d',df.index.freqstr)[0]
        print(df.index[0].freqstr)
        for k,tagName in zip(range(len(df)),list(df.columns)):
             tmpRes = self.utils.fitSingle(df.iloc[:,[k]],func=func,**kwargs,plotYes=plotYes)
             res[tagName] = [tmpRes[0],tmpRes[1],tmpRes[2],
                            1/tmpRes[1]/float(period),tmpRes[0]+tmpRes[2]]
        res  = pd.DataFrame(res,index = ['a','b','c','tau(s)','T0'])
        return res

class ConfigDashRealTime(ConfigDashTagUnitTimestamp):
    def __init__(self,confFolder,connParameters):
        import glob
        from dorianUtils.utilsD import DataBase
        super().__init__(folderPkl='',confFolder=confFolder)
        self.dataBaseUtils  = DataBase()
        self.connParameters = connParameters

    def _getPLC_ColName(self):
        l = self.dfPLC.columns
        v = l.to_frame()
        unitCol = ['unit' in k.lower() for k in l]
        descriptName = ['descript' in k.lower() for k in l]
        tagName = ['tag' in k.lower() for k in l]
        return [list(v[k][0])[0] for k in [unitCol,descriptName,tagName]]

    def connectToDB(self):
        return self.dataBaseUtils.connectToPSQLsDataBase(self.connParameters)

    def processRawData(self,df,rs='1s',applyMethod='mean'):
        if df.duplicated().any():
            df = df.drop_duplicates()
            print("attention il y a des doublons dans la base de donnees Jules")
        df['value'] = pd.to_numeric(df['value'],errors='coerce')
        df = df.sort_values(by=['timestampz','tag'])
        df['timestampz'] = df.timestampz.dt.tz_convert('Europe/Paris')
        df = df.pivot(index="timestampz", columns="tag", values="value")
        df = eval("df.resample('100ms').ffill().ffill().resample(rs).apply(np." + applyMethod + ")")
        return df

    def realtimeDF(self,preSelGraph,rs='1s',applyMethod='mean',timeWindow=60*60*2):
        if rs=='auto':rs = '{:.0f}'.format(max(1,timeWindow//1000)) + 's'
        preSelGraph = self.usefulTags.loc[preSelGraph]
        conn = self.connectToDB()
        df   = self.dataBaseUtils.readSQLdataBase(conn,preSelGraph.Pattern,secs=timeWindow)
        if df.duplicated().any():
            df = df.drop_duplicates()
            print("attention il y a des doublons dans la base de donnees Jules")
        df['value'] = pd.to_numeric(df['value'],errors='coerce')
        df = df.sort_values(by=['timestampz','tag'])
        df['timestampz'] = df.timestampz.dt.tz_convert('Europe/Paris')
        df = df.pivot(index="timestampz", columns="tag", values="value")
        df = eval("df.resample('100ms').ffill().ffill().resample(rs).apply(np." + applyMethod + ")")
        conn.close()
        return df

    def realtimeTagsDF(self,tags,timeWindow=60*60*2,rs='1s',applyMethod='mean',timeRange=None):
        if rs=='auto':rs = '{:.0f}'.format(max(1,timeWindow//1000)) + 's'
        else :
            conn = self.connectToDB()
            df   = self.dataBaseUtils.readSeveralTagsSQL(conn,tags,secs=timeWindow,timeRange=timeRange)
            df   = self.processRawData(df,rs=rs,applyMethod=applyMethod)
        return df

    def doubleMultiUnitGraph(self,df,tags1,tags2,axSP=0.05):
        dictdictGroups={'graph1':{t:t for t in tags1},'graph2':{t:t for t in tags2}}
        fig = self.utils.multiUnitGraphSubPlots(df,dictdictGroups,axisSpace=axSP)
        hs=0.002
        for y in range(1,len(tags1)+1):
            fig.layout['yaxis1'+str(y)].domain=[0.5+hs,1]
        for y in range(1,len(tags2)+1):
            fig.layout['yaxis2'+str(y)].domain=[0,0.5-hs]
        fig.update_layout(xaxis_showticklabels=False)
        fig.update_yaxes(title_text='',showticklabels=False)
        fig.update_yaxes(showgrid=False)
        fig.update_xaxes(matches='x')
        return fig

class ConfigDashSpark(ConfigDashTagUnitTimestamp):
    try :
        from pyspark.sql import SparkSession
        from dorianUtils.sparkUtils.sparkDfUtils import SparkDfUtils
        import findspark

        def __init__(self,sparkData,sparkConfFile,confFolder,**kwargs):
            ConfigDashTagUnitTimestamp.__init(self,folderPkl=None,confFolder=confFolder,**kwargs)

            self.cfgSys     = self._getSysConf(sparkConfFile)
            self.spark      = self.__createSparkSession()
            self.sparkConf  = self.spark.sparkContext.getConf().getAll()

            self.sparkData  = sparkData

            self.sdu   = SparkDfUtils(self.spark)
        # ======================================================================
        #                         private functions
        # ======================================================================
        def _getSysConf(self,scriptFile):
            def toString(b):
                s = str(b).strip()
                if (s.startswith("b'")): s = s[2:].strip()
                if (s.endswith("'")): s = s[:-1].strip()
                return s
            import subprocess
            conf = None
            process = subprocess.Popen(scriptFile.split(),
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    shell=True,
                                    executable="/bin/bash")
            stdout, stderr = process.communicate()
            stderr = toString(stderr)
            if (len(stderr) < 3):
                conf = {}
                stdout = toString(stdout)
                stdout = stdout.replace("\\n", '\n')
                stdout = stdout.split('\n')
                for kv in stdout:
                    kv = kv.split('=', 1)
                    if (len(kv)==2 and not ' ' in kv[0] and kv[0] != '_'):
                        kv[0] = kv[0].strip()
                        conf.update({kv[0]: kv[1]})
            return conf

        def _getSparkParams(self):
            def clean(line, sep, index):
                value = line.split(sep, 1)[index]
                value = value.strip()
                if (value.startswith('"') or value.startswith("'")): value = value[1:]
                if (value.endswith('"') or value.endswith("'")): value = value[:-1]
                return value
            sparkCmd = self.cfgSys.get("SPARK_CMD")
            if (sparkCmd == None):
                print("")
                print("ERROR: 'SPARK_CMD' environment variable is missing !")
                print("")
                sys.exit(0)
            lines = sparkCmd.split("--")
            params = []
            for line in lines:
                line = line.strip()
                # name
                if (line.startswith("name ")):
                    params.append({ "type": "name", "key": "name", "value": clean(line, ' ', 1) })
                # master
                elif (line.startswith("master ")):
                    params.append({ "type": "master", "key": "master", "value": clean(line, ' ', 1) })
                # deploy-mode
                elif (line.startswith("deploy-mode ")):
                    params.append({ "type": "conf", "key": "spark.submit.deployMode", "value": clean(line, ' ', 1) })
                # conf
                elif (line.startswith("conf ")):
                    kvpair = line.split(' ', 1)[1]
                    params.append({ "type": "conf", "key": clean(kvpair, '=', 0), "value": clean(kvpair, '=', 1) })
                # packages
                elif (line.startswith("packages ")):
                    libs = [lib.strip() for lib in line[8:].split(',')]
                    libs = ','.join(libs)
                    params.append({ "type": "packages", "key": "spark.jars.packages", "value": libs })
                # jars
                elif (line.startswith("jars ")):
                    libs = [lib.strip() for lib in line[4:].split(',')]
                    libs = ','.join(libs)
                    params.append({ "type": "jars", "key": "spark.jars", "value": libs })
                # driver-class-path
                elif (line.startswith("driver-class-path ")):
                    libs = [lib.strip() for lib in line[17:].split(',')]
                    libs = ','.join(libs)
                    params.append({ "type": "driver-class-path", "key": "spark.driver.extraClassPath", "value": libs })
            return params

        def __createSparkSession(self):
            params = self._getSparkParams()
            # Get the builder
            builder = SparkSession.builder
            # Set name
            for param in params:
                if (param.get("type")=="name"):
                    builder = builder.appName(param.get("value"))
                    break
            # Set master
            for param in params:
                if (param.get("type")=="master"):
                    builder = builder.master(param.get("value"))
                    break
            # Set conf params - Include "spark.submit.deployMode" (--deploy-mode) and "spark.jars.packages" (--packages)
            for param in params:
                if (param.get("type")=="conf"):
                    builder = builder.config(param.get("key"), param.get("value"))
            # Set packages
            for param in params:
                if (param.get("type")=="packages"):
                    builder = builder.config(param.get("key"), param.get("value"))
            # Set jars libs
            for param in params:
                if (param.get("type")=="jars"):
                    builder = builder.config(param.get("key"), param.get("value"))
            # Set driver-class-path libs
            for param in params:
                if (param.get("type")=="driver-class-path"):
                    builder = builder.config(param.get("key"), param.get("value"))
            # Enable HIVE support
            ehs = self.cfgSys.get("ENABLE_HIVE_SUPPORT")
            if (ehs != None):
                ehs = ehs.strip().lower()
                if (ehs == "true"):
                    builder = builder.enableHiveSupport()
                    print(" - Spark HIVE support enabled")
            # Create the Spark session
            spark = builder.getOrCreate()
            spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
            return spark
        # ======================================================================
        #                         function on dataframe
        # ======================================================================

        def getPartitions(self,timeRange):
            from dateutil import parser

            def createSinglePartition(curDay,hours):
                vars = ['year','month','day']
                # partoch=[l.upper() + "=" + str(eval('curDay.'+ l)) for l in vars] # doesn't work !!
                partoch = []
                for k in vars:partoch.append(k.upper() + "=" + str(eval('curDay.'+k)))
                tmp = '{' + ','.join([str(k) for k in range(hours[0],hours[1])]) + '}'
                partoch.append('HOUR=' + tmp)
                partoch = '/'.join(partoch)
                return partoch

            t0,t1     = [parser.parse(k) for k in timeRange]
            lastDay=(t1-t0).days
            if lastDay==0:
                curDay=t0
                partitions = [createSinglePartition(t0,[t0.hour,t1.hour])]
            else :
                curDay=t0
                partitions = [createSinglePartition(t0,[t0.hour,24])]
                for k in range(1,lastDay):
                    curDay=t0+dt.timedelta(days=k)
                    curDay=curDay-dt.timedelta(hours=curDay.hour,minutes=curDay.minute)
                    partitions.append(createSinglePartition(curDay,[0,24]))
                curDay=t1
                partitions.append(createSinglePartition(t1,[0,t1.hour]))
            return partitions

        def getPartitions_v0(self,timeRange):
            t0,t1 = [parser.parse(k) for k in timeRange]
            yearRange = str(t0.year)
            monthRange="{" + str(t0.month) + "," + str(t1.month) + "}"
            dayRange="{" + str(t0.day) + "," + str(t1.day) + "}"
            hourRange="[" + str(t0.hour) + "," + str(t1.hour) + "]"
            # hourRange="{" + str(t0.hour) + "," + str(t1.hour) + "}"
            partitions = {
                        "YEAR" : yearRange,
                        "MONTH" : monthRange,
                        "DAY" : dayRange,
                        "HOUR" : hourRange,
            }
            partitions = [k + "=" + v for k,v in partitions.items()]
            return '/'.join(partitions)

        def loadSparkTimeDF(self,timeRange,typeData="EncodedData"):
            ''' typeData = {
            encoded data : "EncodedData"
            aggregated data : "AggregatedData`"
            populated data :"PopulatedData"
            }
            '''
            df = self.sdu.loadParquet(inputDir=self.sparkData+typeData+"/",partitions=self.getPartitions(timeRange))
            df = self.sdu.organizeColumns(df, columns=["YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND"], atStart=True)
            timeSQL = "TIMESTAMP_UTC" + " BETWEEN '" + timeRange[0] +"' AND '" + timeRange[1] +"'"
            return df.where(timeSQL)

        def getSparkTU(self,df,listTags):
            dfs=[]
            if not isinstance(listTags,list):listTags=[listTags]
            for tag in listTags:
                print(tag)
                df2 = df.where("TAG == "+ "'" + tag + "'")
                df3 = df2.select('TAG','VALUE','TIMESTAMP_UTC')
                dfs.append(df3.toPandas())
            pdf = pd.concat(dfs).sort_values(by=['TIMESTAMP_UTC','TAG'])
            return pdf

        def getSparkTU_v2(self,df,listTags):
            dfs=self.sdu.createDataFrame(schema=[["TAG","string"],["VALUE","double"],["TIMESTAMP_UTC","timestamp"]])

            if not isinstance(listTags,list):listTags=[listTags]
            for tag in listTags:
                print(tag)
                df2 = df.where("TAG == "+ "'" + tag + "'").select('TAG','VALUE','TIMESTAMP_UTC')
                dfs=dfs.unionByName(df2)
            pdf = dfs.toPandas().sort_values(by=['TIMESTAMP_UTC','TAG'])
            return pdf
            # return dfs
    except:
        print('not possible to load : module pyspark ')

class ConfigFilesRealTimeCSV(ConfigDashTagUnitTimestamp):
    def __init__(self,folderRT,confFolder):
        ConfigDashTagUnitTimestamp.__init__(self,folderRT,confFolder,folderFig='')
        self.folderRT = folderRT
        self.listParkedDays=[k for k in os.listdir(folderRT) if re.search('\d{4}-\d{2}-\d{2}',k)]
        self.mbUtils = Modebus_utils()
        self.listParkTags=self.getListTagsParked()

    def findUnitTagName(self,t,sep=['-']):
        listUnits = [k for s in list(self.utils.phyQties.values()) for k in s ]
        listUnits = self.utils.combineUnits(self.utils.unitMag,listUnits,'')
        listUnitsPat=[l + k + l for k in listUnits for l in seps]
        return [k for k in listUnitsPat if re.search(k,t)]

    def getListTagsParked(self):
        if len(self.listParkedDays)>0:
            return [k[:-4] for k in os.listdir(self.folderRT+self.listParkedDays[0])]
        else : return []

    def getTagsTU(self,patTag,units=None,onCol=None,case=False,cols=None,ds=True):
        return [k for k in self.dfPLC.TAG if re.search(patTag,k)]

    def realtimeTagsDF(self,listTags,**kwargs):
        return self.mbUtils.readTagsRealTime(self.folderRT,listTags,**kwargs)



from mpl_toolkits.axes_grid.parasite_axes import SubplotHost
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
from dateutil import parser
import sys
from dorianUtils.utilsD import Utils
utils=Utils()

def make_patch_spines_invisible(ax):
    ax.set_frame_on(True)
    ax.patch.set_visible(False)
    for sp in ax.spines.values():
        sp.set_visible(False)

def plotmtpaxis(x,y,sc=0.2,rm=0.65,lm=1,xlab='xlab',ylabs=[],markers=[],figSize=[8,8]):
    '''plot multiple axis
    x is a list
    y are list of lists'''
    fig, host = plt.subplots()
    fig.subplots_adjust(right=rm)
    fig.set_size_inches(figSize[0],figSize[1])
    palette = plt.get_cmap('jet',len(y))
    tkw = dict(size=6, width=1.5)
    if not markers:
        markers = ['o' for k in range(len(y))]
    if not ylabs:
        ylabs = ['ylab ' + str(k) for k in range(len(y))]
    #===================== host =============================
    host.plot(x, y[0],color=palette(0), linestyle="--", marker=markers[0], label=ylabs[0])
    host.set_xlabel(xlab)
    host.set_ylabel(ylabs[0])
    host.yaxis.label.set_color(palette(0))
    host.tick_params(axis='x', **tkw)
    host.tick_params(axis='y', colors=palette(0), **tkw)
    host.spines["left"].set_color(palette(0))
    pars = []
    #============================= parasite axes =============================r
    for k in range(1,len(y)):
        # par[k] =
        par= host.twinx()
        par.spines["right"].set_position(("axes", 1+(k-1)*sc))
        make_patch_spines_invisible(par)
        par.spines["right"].set_visible(True)
        par.spines["right"].set_color(palette(k))
        par.plot(x, y[k], color=palette(k),linestyle="--",marker=markers[k], label=ylabs[k])
        par.set_ylabel(ylabs[k])
        par.yaxis.label.set_color(palette(k))
        par.tick_params(axis='y', colors=palette(k), **tkw)
        # pars[k]=par
        # del par
    return fig

class multiY():
    #credits: http://matplotlib.sourceforge.net/examples/pylab_examples/multiple_yaxis_with_spines.html
    def __init__(self, height, width, X, LY, Xlabel, LYlabel, linecolor,
    set_marker, set_linestyle, fontsize, set_markersize,
    set_linewidth, set_mfc, set_mew, set_mec):

        self.X = X

        fig = plt.figure(figsize=(height, width))

        self.host = SubplotHost(fig, 111)

        plt.rc("font", size=fontsize)

        self.host.set_xlabel(Xlabel)
        self.host.set_ylabel(LYlabel)
        p1, = self.host.plot(X, LY, color=linecolor, marker=set_marker, ls=set_linestyle, ms=set_markersize, lw=set_linewidth, mfc=set_mfc, mew=set_mew, mec=set_mec)

        fig.add_axes(self.host)

        self.host.axis["left"].label.set_color(p1.get_color())
        self.host.tick_params(axis='y', color=p1.get_color())

    def parasites(self, set_offset, PY, PYlabel, side, Plinecolor,
    Pset_marker, Pset_linestyle, Pset_markersize, Pset_linewidth, Pset_mfc, Pset_mew, Pset_mec):
        par = self.host.twinx()
        par.axis["right"].set_visible(False)
        offset = (set_offset, 0)
        new_axisline = par.get_grid_helper().new_fixed_axis

        par.axis["side"] = new_axisline(loc=side, axes=par, offset=offset)

        par.axis["side"].label.set_visible(True)
        par.axis["side"].set_label(PYlabel)

        p2, = par.plot(self.X, PY,color=Plinecolor, marker=Pset_marker, ls=Pset_linestyle, ms=Pset_markersize, lw=Pset_linewidth, mfc=Pset_mfc, mew=Pset_mew, mec=Pset_mec)

        par.axis["side"].label.set_color(p2.get_color())
        par.tick_params(axis='y', colors=p2.get_color())

def demo_multiY():
#data
    x = (1, 3, 4, 6, 8, 9, 12)
    y1 = (0, 1, 2, 2, 4, 3, 2)
    y2 = (0, 3, 2, 3, 6, 4, 5)
    y3 = (50, 40, 40, 30, 20, 22, 10)
    y4 = (0.2, 0.5, 0.6, 0.9, 2, 5, 2)
    y5 = (2, 0.5, 0.9, 9, 12, 15, 12)
    y6 = (200, 500, 900, 900, 120, 150, 120)

    #height, width, x-data, y-datam, X-label, Y-label, line color, markerstyle, linestyle,
    # fontsize, markersize, linewidth, marker face color, marker edge with, marker edgecolor
    aa = multiY(8, 8, x, y1, "X-label", "Y-label", "r", "o", "None", 16, 8, 2, "none", "1", "r")
    #offset, y-data, Y-label, side, color, marker edge with, marker edgecolor
    aa.parasites(0, y2, "Parasite2", "right", "g", ">", "--", 10, 1, "none", "1", "g")
    aa.parasites(-60, y3, "Parasite3", "left", "b", "<", "-", 10, 1, "none", "1", "b")
    # aa.parasites(50, y4, "Parasite4", "right", "m", "s", "-.", 10, 1, "none", "1", "m")
    # aa.parasites(-100, y5, "Parasite5", "left", "y", "d", "none", 10, 1, "none", "1", "y")
    # aa.parasites(100, y6, "Parasite6", "right", "k", "*", "--", 10, 1, "none", "1", "k")

    #adjust the plot
    plt.subplots_adjust(left=0.20, bottom=0.15, right=0.78, top=0.92, wspace=0.05, hspace=0)
    plt.show()

def multiYmpl(df,**kwargs):
    listNames = df.columns
    listY = [df[k] for k in listNames]
    fig = plotmtpaxis(df.index,listY,**kwargs,ylabs=list(listNames))
    return fig

def demo_multiYmpl():
    df = px.data.stocks()
    df.index=[parser.parse(k) for k in df.date]
    del df['date']
    fig = multiYmpl(df.iloc[::10,:],sc=0.1,rm=0.7,figSize=[16,13],xlab='time (s)')
    plt.show(block=False)

def plotMultiUnitSmallPower(fileNb=0,skip=5,folder =None ,**kwargs):
    if not folder :
        folder = '/home/dorian/sylfen/exploreSmallPower/dataExported/'
    lsfiles = utils.get_listFilesPkl(folder,'.txt')
    print(utils.listWithNbs(lsfiles))
    df = pd.read_csv(folder + lsfiles[fileNb],index_col = 0,parse_dates=True)
    fig = multiYmpl(df.iloc[::skip,:],xlab='time',**kwargs)
    plt.show(block=False)



import pandas as pd, numpy as np, pickle, re, time, datetime as dt,glob
from datetime import timezone
import subprocess as sp, os
import plotly.graph_objects as go
import plotly.express as px
from plotly.validators.scatter.line import ShapeValidator
from plotly.validators.scatter.marker import SymbolValidator
from plotly.validators.scatter.line import DashValidator

def dimensions_pictures_screen(diag_inch=15,resolution=[1920,1080]):
    res={}
    res['a_ratio'] = 16/9
    res['resolution'] = resolution
    res['inch2cm'] = 2.54
    res['screen_w'] = np.cos(np.arctan(1/res['a_ratio']))*diag_inch
    res['screen_h'] = np.sin(np.arctan(1/res['a_ratio']))*diag_inch
    dpi_w=resolution[0]/res['screen_w']
    dpi_h=resolution[1]/res['screen_h']
    res['dpi']=np.mean([dpi_h,dpi_w])
    return res

class Utils:
    def __init__(self):
        self.confDir=os.path.dirname(os.path.realpath(__file__)) + '/conf'
        self.phyQties = self.df2dict(pd.read_csv(self.confDir+ '/units.csv'))
        self.unitMag = ['u','m','c','d','','da','h','k','M']
        self.buildNewUnits()
        raw_symbols = pd.Series(SymbolValidator().values[2::3])
        self.raw_symbols = list(pd.concat([raw_symbols[::4],raw_symbols[1::4],raw_symbols[2::4]]))
        self.listLines = 40*DashValidator().values[:-1]
        self.lineshapes = ShapeValidator().values
        allcolors=px.colors.qualitative.Plotly.copy()
        allcolors+=px.colors.qualitative.Dark24.copy()+px.colors.qualitative.Light24
        allcolors+=px.colors.qualitative.Alphabet
        allcolors+=px.colors.qualitative.Set1+px.colors.qualitative.Pastel1
        allcolors+=px.colors.qualitative.Antique
        self.colors_mostdistincs = allcolors
        self.styles = ['lines+markers','markers','stairs','lines']
        self.colorPalettes=pd.read_pickle(self.confDir + '/palettes.pkl')

        from PIL import Image # new import
        self.sylfenlogo  = Image.open(self.confDir +  '/logo_sylfen.png')

    # ==========================================================================
    #                               DEBUG
    # ==========================================================================

    def printCTime(self,start,entete='time laps' ):
        return entete + ' : {:.2f} seconds'.format(time.time()-start)

    def printListArgs(self,*args):
        for a in args :
            print(' ------- ')
            print(a)
        print('============ args finished =================')

    # ==========================================================================
    #                               SYSTEM
    # ==========================================================================
    def read_csv_datetimeTZ(self,filename,overwrite=False,**kwargs):
        start   = time.time()
        print("============================================")
        print('reading of file',filename)
        df      = pd.read_csv(filename,**kwargs,names=['tag','value','timestampUTC'])
        self.printCTime(start)
        start = time.time()
        print("============================================")
        print("parsing the dates : ",filename)
        df.timestampUTC=pd.to_datetime(df.timestampUTC,utc=True)# convert datetime to utc
        df['value'] = pd.to_numeric(df['value'],errors='coerce')
        self.printCTime(start)
        print("============================================")
        return df

    def convert_csv2pkl(self,folderCSV,folderPKL,overwrite=False):
        try :
            listFiles=self.get_listFilesPkl(folderCSV,'.csv')
            if not overwrite:
                listFiles = [f for f in listFiles if not f[:-4]+'.pkl' in folderPKL]
                for filename in listFiles:
                    df=self.read_csv_datetimeTZ(folderCSV + filename)
                    with open(folderPKL + filename[:-4] + '.pkl' , 'wb') as handle:
                        pickle.dump(df, handle, protocol=pickle.HIGHEST_PROTOCOL)
        except:
            print('no csv files in directory : ',folderCSV)

    def get_listFilesPkl(self,folderName=None,ext='.pkl'):
        if not folderName :folderName = os.getcwd()
        listFiles = sp.check_output('cd ' + '{:s}'.format(folderName) + ' && ls *' + ext,shell=True)
        listFiles=listFiles.decode().split('\n')[:-1]
        return listFiles

    def get_listFilesPklV2(self,folderName=None,pattern='*.pkl'):
        if not folderName :folderName = os.getcwd()
        listfiles = glob.glob(folderName+pattern)
        listfiles.sort()
        return listfiles

    def skipWithMean(self,df,windowPts,idxForMean=None,col=None):
        ''' compress a dataframe by computing the mean around idxForMean points'''
        if not col :
            col = [k for k in range(len(df.columns))]
        if not idxForMean :
            idxForMean = list(range(windowPts,len(df),windowPts))
        ll = [df.iloc[k-windowPts:k+windowPts+1,col].mean().to_frame().transpose()
                for k in idxForMean]
        dfR = pd.concat(ll)
        dfR.index = df.index[idxForMean]
        return dfR

    def slugify(self,value, allow_unicode=False):
        import unicodedata,re
        """
        Taken from https://github.com/django/django/blob/master/django/utils/text.py
        Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
        dashes to single dashes. Remove characters that aren't alphanumerics,
        underscores, or hyphens. Convert to lowercase. Also strip leading and
        trailing whitespace, dashes, and underscores.
        """
        value = str(value)
        if allow_unicode:value = unicodedata.normalize('NFKC', value)
        else:value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value.lower())
        return re.sub(r'[-\s]+', '-', value).strip('-_')

    def is_dst(self,t=None, timezone="UTC"):
        if t is None:t = dt.utcnow()
        timezone = pytz.timezone(timezone)
        timezone_aware_date = timezone.localize(t, is_dst=None)
        return timezone_aware_date.tzinfo._dst.seconds != 0

    def findDateInFilename(self,filename,formatDate='\d{4}-\d{2}-\d{2}'):
        if '/' in filename:filename = filename.split('/')[-1]
        # print('filename:',filename)
        tmax = re.findall(formatDate,filename)[0].split('-')# read the date of the last file in the folder
        # print('tmax:',tmax)
        tmax = dt.datetime(int(tmax[0]),int(tmax[1]),int(tmax[2]))
        return tmax

    def writeLog(self,logFile,msg,newBlock=False):
        goLine = '================================================================'
        with open(logFile, 'a') as f:
            if newBlock:f.write(goLine + '\n')
            f.write(msg + '\n')
            if newBlock:f.write(goLine + '\n')

    def getSysConf(self, scriptFile):
        def toString(b):
            s = str(b).strip()
            if (s.startswith("b'")): s = s[2:].strip()
            if (s.endswith("'")): s = s[:-1].strip()
            return s
        conf = None
        with open(scriptFile) as f: s= ''.join(f.readlines())+'\nenv'
        process = sp.Popen(s,stdout=sp.PIPE,stderr=sp.PIPE,shell=True,
                                    executable="/bin/bash")
        stdout, stderr = process.communicate()

        stderr = toString(stderr)
        if (len(stderr) < 3):
            conf = {}
            stdout = toString(stdout)
            stdout = stdout.replace("\\n", '\n')
            stdout = stdout.split('\n')
            for kv in stdout:
                kv = kv.split('=', 1)
                if (len(kv)==2 and not ' ' in kv[0] and kv[0] != '_'):
                    kv[0] = kv[0].strip()
                    conf.update({kv[0]: kv[1]})
        else : print(stderr)

        return conf

    # ==========================================================================
    #                           PHYSICS
    # ==========================================================================
    def buildNewUnits(self):
        self.phyQties['vitesse'] = self.combineUnits(self.phyQties['distance'],self.phyQties['temps'])
        self.phyQties['mass flow'] = self.combineUnits(self.phyQties['masse'],self.phyQties['temps'])
        tmp = self.combineUnits(['','N'],self.phyQties['volume'],'')
        self.phyQties['volumetric flow'] = self.combineUnits(tmp,self.phyQties['temps'])
        self.phyQties['conducitivité'] = self.combineUnits(self.phyQties['conductance'],self.combineUnits(self.unitMag,self.phyQties['distance'],''))

    def combineUnits(self,units1,units2,oper='/'):
        return [x1 + oper + x2 for x2 in units2 for x1 in units1]

    def detectUnit(self,unit):
        phId = ''
        for phyQt in self.phyQties.keys():
            # listUnits = [x1+x2 for x2 in self.phyQts[phyQt] for x1 in self.unitMag]
            listUnits = self.combineUnits(self.unitMag,self.phyQties[phyQt],'')
            if unit in listUnits : phId = phyQt
        return phId

    def detectUnits(self,listUnits,check=0):
        tmp = [self.detectUnit(unit) for unit in listUnits]
        if check :
            listUnitsDf = pd.DataFrame()
            listUnitsDf['units'] = listUnits
            listUnitsDf['grandeur'] = tmp
            return listUnitsDf
        else :
            return tmp

    def lowpass(self,x,alpha=0.01):
        newx=x.copy()
        for k in range(1,len(x)):
            newx[k] = (1-alpha)*newx[k-1] + alpha*newx[k]
        return newx

    # ==========================================================================
    #                       lIST AND DICTIONNARIES
    # ==========================================================================
    def df2dict(self,df):
        return {df.columns[k] : list(df.iloc[:,k].dropna()) for k in range(len(df.columns))}

    def linspace(self,arr,numElems):
        idx = np.round(np.linspace(0, len(arr) - 1, numElems)).astype(int)
        return list([arr[k] for k in idx])

    def flattenList(self,l):
        return [item for sublist in l for item in sublist]

    def flattenDict(self,ld):
        finalMap = {}
        for d in ld:finalMap.update(d)
        return finalMap

    def removeNaN(self,list2RmNan):
        tmp = pd.DataFrame(list2RmNan)
        return list(tmp[~tmp[0].isna()][0])

    def sortIgnoCase(self,lst):
        df = pd.DataFrame(lst)
        return list(df.iloc[df[0].str.lower().argsort()][0])

    def dfcolwithnbs(self,df):
        a = df.columns.to_list()
        coldict=dict(zip(range(0,len(a)),a))
        coldict
        return coldict

    def listWithNbs(self,l,withDF=False):
        if withDF:return pd.DataFrame(l)
        else : return [str(i) + ' : '+ str(k) for i,k in zip(range(len(l)),l)]

    def dspDict(self,dict,showRows=1):
        '''display dictionnary in a easy readable way :
        dict_disp(dict,showRows)
        showRows = 1 : all adjusted '''
        maxLen =max([len(v) for v in dict])
        for key, value in dict.items():
            valToShow = value
            if showRows == 0:
                rowTxt = key.ljust(maxLen)
            if showRows == 1:
                if len(key)>8:
                    rowTxt = (key[:8]+'..').ljust(10)
                else:
                    rowTxt = key.ljust(10)
            if showRows==-1:
                rowTxt      = key.ljust(maxLen)
                valToShow   = type(value)
            if showRows==-2:
                rowTxt      = key.ljust(maxLen)
                valToShow   = value.shape
            print(colored(rowTxt, 'red', attrs=['bold']), ' : ', valToShow)

    def regExpNot(self,regexp):
        if regexp[:2] == '--': regexp = '^((?!' + regexp[2:] + ').)*$'
        return regexp

    def uniformListStrings(self,l):
        newList=[]
        for k in l:
            li=[m.start(0) for m in re.finditer('\w',k)]
            newList.append(k[li[0]:li[-1]+1].capitalize())
        return newList

    # ==========================================================================
    #                                   DATAFRAMES
    # ==========================================================================
    def loads_pickle(self,file_pkl):
        f = open(file_pkl,'rb')
        objs = []
        while 1:
            try:
                objs.append(pickle.load(f))
            except EOFError:
                break
        return objs

    def combineFilter(self,df,columns,filters):
        cf  = [df[col]==f for col,f in zip(columns,filters)]
        dfF = [all([cfR[k] for cfR in cf]) for k in range(len(cf[0]))]
        return df[dfF]

    def pivotDataFrame(self,df,colTagValTS=None,resampleRate='60s',applyMethod='nanmean'):
        if not colTagValTS : colTagValTS = [0,1,2]
        colTagValTS = df.columns[colTagValTS]
        listTags = list(df[colTagValTS[0]].unique())
        t0 = df[colTagValTS[2]].min()
        dfOut = pd.DataFrame()
        for tagname in listTags:
            dftmp = df[df[colTagValTS[0]]==tagname]
            dftmp = dftmp.set_index(colTagValTS[2])
            dftmp = eval('dftmp.resample(resampleRate,origin=t0).apply(np.' + applyMethod + ')')
            dfOut[tagname] = dftmp[colTagValTS[1]]

        # dfOut=dfOut.fillna(method='ffill')
        return dfOut

    def dictdict2df(self,dictdictGroups):
        dfGroups=pd.DataFrame.from_dict(dictdictGroups)
        dfGroups['tag']=dfGroups.index
        dfGroups=dfGroups.melt(id_vars='tag')
        dfGroups=dfGroups.dropna().set_index('tag')
        dfGroups.columns=['group','subgroup']
        return dfGroups

    def export2excel(df,filepath):
        if isinstance(df.index,pd.core.indexes.datetimes.DatetimeIndex):
            df.index=[k.isoformat() for k in df.index]
        df.to_excel(filepath)

    def popup_dfexcel(self,df):
        if isinstance(df.index,pd.core.indexes.datetimes.DatetimeIndex):
            df.index=[k.isoformat() for k in df.index]
        df.to_excel('/tmp/test.xlsx')
        sp.Popen(['libreoffice','/tmp/test.xlsx'])

    def showdf_as_table(self,df):
        fig = go.Figure(data=[go.Table(
        header=dict(values=list(df.columns),
        fill_color='paleturquoise',
        align='left'),
        cells=dict(values=[df[k] for k in df.columns],
        fill_color='lavender',
        align='left'))
        ])
        fig.show()
        return fig

    def dict2df(self,d):
        return pd.concat([pd.Series(d[v],name=v) for v in d.keys()],axis=1)
    # ==========================================================================
    #                           GRAPHICS
    # ==========================================================================
    def sample_colorscale(self,N,colorscale='jet'):
        return px.colors.sample_colorscale(colorscale,np.linspace(0,1,N))

    def updateColorMap(self,fig,listCols=None):
        if listCols is None:
            listCols = self.colors_mostdistincs
        if isinstance(listCols,str):
            listCols = self.sample_colorscale(len(fig.data)+1,colorscale=listCols)
        k,l=0,0
        listYaxis = [k for k in fig._layout.keys() if 'yax' in k]
        if len(listYaxis)>1:
            for yax in listYaxis :
                k+=1
                fig.layout[yax]['title']['font']['color'] = listCols[k]
                fig.layout[yax]['tickfont']['color'] = listCols[k]
        for d in fig._data :
            l+=1
            if 'marker' in d.keys():
                d['marker']['color']=listCols[l]
            if 'line' in d.keys():d['line']['color']=listCols[l]
        return fig

    def optimalGrid(self,n):
        if n==1:return [1,1]
        elif n==2:return [2,1]
        elif n==3:return [3,1]
        elif n==4:return [2,2]
        elif n==5:return [3,2]
        elif n==6:return [3,2]
        elif n==7:return [4,2]
        elif n==8:return [4,2]
        elif n==9:return [3,3]
        elif n==10:return [5,2]

    def rowsColsFromGrid(self,n,grid):
        i,rows,cols=0,[],[]
        idxMin=grid.index(min(grid))
        while i<n+1:
            rows.append(i%min(grid)+1)
            cols.append(i//min(grid)+1)
            # print(i,rows[i],cols[i])
            i+=1
        if idxMin==0:return rows,cols
        else:return cols,rows

    def customLegend(self,fig, nameSwap,breakLine=None):
        if not isinstance(nameSwap,dict):
            print('not a dictionnary, there may be wrong assignment')
            namesOld = [k.name  for k in fig.data]
            nameSwap = dict(zip(namesOld,nameSwap))
        for i, dat in enumerate(fig.data):
            for elem in dat:
                if elem == 'name':
                    newName = nameSwap[fig.data[i].name]
                    if isinstance(breakLine,int):
                        newName = '<br>s'.join([newName[k:k+breakLine] for k in range(0,len(newName),breakLine)])
                    fig.data[i].name = newName
        return fig

    def makeFigureName(self,filename,patStop,toAdd):
        idx=filename.find(patStop)
        f=filename[:idx]
        f=re.sub('[\./]','_','_'.join([f]+toAdd))
        return f

    def figureName(self,params,joinCara=',',egal='=',patAfterList=None):
        listParams=[]
        if not patAfterList : patAfterList=['']*len(params)
        for (k,v),u in zip(params.items(),patAfterList):
            tmp = ''
            if isinstance(v,int):tmp = k + egal + '{:d}'.format(v) + ' ' +u
            if isinstance(v,float):tmp=k + egal + '{:1f}'.format(v) + ' ' +u
            if isinstance(v,str):
                if len(v)>0 : tmp= k + egal + v + ' ' +u
            if not not tmp:listParams.append(tmp)
        return joinCara.join(listParams)

    def buildTimeMarks(self,t0,t1,nbMarks=8,fontSize='12px'):
        maxSecs=int((t1-t0).total_seconds())
        listSeconds = [int(t) for t in np.linspace(0,maxSecs,nbMarks)]
        dictTimeMarks = {k : {'label':(t0+dt.timedelta(seconds=k)).strftime('%H:%M'),
                                'style' :{'font-size': fontSize}
                                } for k in listSeconds}
        return dictTimeMarks,maxSecs

    def getAutoYAxes(self,N,xrange,inc):
        sides =['left','right']*6 # alterne
        # print(xrange)
        positions  = self.flattenList([[xrange[0]-k*inc,xrange[1]+k*inc] for k in range(6)])
        return sides[:N],positions[:N]

    def getLayoutMultiUnit(self,dictGroups,colormap='jet',axisSpace=0.05):
        dfGroups = pd.DataFrame.from_dict(dictGroups,orient='index',columns=['group'])
        groups=dfGroups.group.unique()

        yaxes=['yaxis'] + ['yaxis'+str(k) for k in range(2,len(groups)+1)]
        minx = (len(groups)-1)//2*axisSpace
        xdomain =[minx,1-minx]
        sides,positions = self.getAutoYAxes(len(groups),xdomain,inc=axisSpace)
        colors=self.sample_colorscale(len(groups),colormap)
        dfGroups['color']=colors[0]
        dfGroups['symbol'] = 'circle'
        dfGroups['line']   = 'solid'
        dictYaxis={}
        yscales=['y'] + ['y'+str(k) for k in range(2,len(groups)+1)]
        for g,c,s,p,y,ys in zip(groups,colors,sides,positions,yaxes,yscales):
            dfGroups.loc[dfGroups.group==g,'color'] = c
            dfGroups.loc[dfGroups.group==g,'yscale'] = ys
            lenGroup=len(dfGroups[dfGroups.group==g])
            dfGroups.loc[dfGroups.group==g,'line']=self.listLines[:lenGroup]
            dfGroups.loc[dfGroups.group==g,'symbol']=self.raw_symbols[:lenGroup]
            if ys=='y' : ov = None
            else : ov = 'y'
            dictYaxis[y] = dict(
                title=g,
                titlefont=dict(color=c),
                tickfont=dict(color=c),
                anchor='free',
                overlaying=ov,
                side=s,
                position=p,
                gridcolor=c
            )
        fig = go.Figure()
        fig.update_layout(dictYaxis)
        fig.update_layout(xaxis=dict(domain=xdomain))
        return fig,dfGroups

    def multiUnitGraph(self,df,dictGroups=None,sizeDots=3,dropna=False):
        if not dictGroups : dictGroups={t:t for t in df.columns}
        fig,dfGroups=self.getLayoutMultiUnit(dictGroups)
        if len(dfGroups.yscale.unique())>12:
            print('too many axes. Please reduce the number of axis')
            return
        for trace in df.columns:
            col=dfGroups.loc[trace,'color']
            y = df[trace]
            if dropna:y = y.dropna()
            fig.add_trace(go.Scatter(
                x=df.index,y=y,name=trace,
                mode="lines+markers",
                yaxis=dfGroups.loc[trace,'yscale'],
                marker=dict(color = col,size=sizeDots,symbol=dfGroups.loc[trace,'symbol']),
                line=dict(color = col,dash=dfGroups.loc[trace,'line'])
                ))
        fig.update_yaxes(showgrid=False)
        return fig

    def getAutoXYAxes(self,n,grid=None,hspace=0.05,minx=0.05,**kwargs):
        from plotly.subplots import make_subplots
        if not grid:grid=self.optimalGrid(n)
        fig = make_subplots(rows=grid[0], cols=grid[1],**kwargs)

        maxx = 1-minx
        if fig.layout['xaxis'].domain[0]==0:
            fig.layout['xaxis'].domain=[minx,fig.layout['xaxis'].domain[1]]
        if fig.layout['xaxis'].domain[1]==1:
            fig.layout['xaxis'].domain=[fig.layout['xaxis'].domain[0],maxx]
        for k in range(2,n+1):
            if fig.layout['xaxis' + str(k)].domain[0]==0:
                fig.layout['xaxis' + str(k)].domain=[minx,fig.layout['xaxis'+ str(k)].domain[1]]
            if fig.layout['xaxis' + str(k)].domain[1]==1:
                fig.layout['xaxis' + str(k)].domain=[fig.layout['xaxis'+ str(k)].domain[0],maxx]
        return fig

    def getLayoutMultiUnitSubPlots(self,dictdictGroups,colormap='jet',axisSpace=0.02,**kwargs):
        dfGroups = self.dictdict2df(dictdictGroups)
        groups = dfGroups.group.unique()
        maxgroups = max([len(dfGroups.groupby('group').get_group(g).subgroup.unique()) for g in groups])
        minx = (maxgroups-1)//2*axisSpace
        # print('minx : ',minx,'===========','maxgroups : ',maxgroups)
        fig=self.getAutoXYAxes(len(groups),minx=minx,**kwargs)

        dfGroups['xaxis']='x'
        dfGroups['yaxis']='y'
        dfGroups['color']='blue'
        dfGroups['symbol']='hexagon-dot'
        dfGroups['line']='solid'

        xaxisNames =['xaxis' + str(k) for k in range(1,len(groups)+1)]
        yaxisNames1 = ['yaxis' + str(k) for k in range(1,len(groups)+1)]
        xscales = ['x' + str(k) for k in range(1,len(groups)+1)]
        yscales1 =['y' + str(k) for k in range(1,len(groups)+1)]

        dictYaxis,dictXaxis = {},{}
        for g,ax,ay1,xs,ys1 in zip(groups,xaxisNames,yaxisNames1,xscales,yscales1):
            # print(ax,' ------ ',ay1,' ------ ',g)
            subgroups=dfGroups[dfGroups.group==g].subgroup.unique()
            colors=self.sample_colorscale(len(subgroups),colormap)
            yaxisNames = [ay1 + str(k) for k in range(1,len(subgroups)+1)]
            yscales = [ys1 + str(k) for k in range(1,len(subgroups)+1)]
            sides,positions = self.getAutoYAxes(len(subgroups),fig.layout[ax].domain,inc=axisSpace)

            dictXaxis[ax] = dict(anchor=ys1+str(1),domain=fig.layout[ax].domain)
            for sg,c,s,p,ys,ay in zip(subgroups,colors,sides,positions,yscales,yaxisNames):
                # print(sg,' ------ ',c,' ------ ',ys)
                # print(dfGroups[(dfGroups.group==g)&(dfGroups.subgroup==sg),'color'])
                filter=(dfGroups.group==g)&(dfGroups.subgroup==sg)
                dfGroups.loc[filter,'color'] = c
                dfGroups.loc[filter,'yaxis'] = ys
                dfGroups.loc[filter,'xaxis'] = xs
                # print(ax,' ------ ',ay1,' ------ ',g)
                lensubgroup=len(dfGroups[dfGroups.subgroup==sg])
                dfGroups.loc[dfGroups.subgroup==sg,'line']=self.listLines[:lensubgroup]
                dfGroups.loc[dfGroups.subgroup==sg,'symbol']=self.raw_symbols[:lensubgroup]
                if ys==ys1+'1' : ov = None
                else : ov = ys1+'1'
                dictYaxis[ay] = dict(
                    title=sg,
                    color=c,
                    anchor='free',
                    domain=fig.layout[ay1].domain,
                    overlaying=ov,
                    side=s,
                    position=p,
                )

        fig.update_layout(dictXaxis)
        fig.update_layout(dictYaxis)
        return fig,dfGroups

    def multiUnitGraphSubPlots(self,df,dictdictGroups,**kwargs):
        fig,dfGroups=self.getLayoutMultiUnitSubPlots(dictdictGroups,**kwargs)
        for trace,g in zip(dfGroups.index,dfGroups.group):
            curTrace =dfGroups[dfGroups.group==g].loc[trace,:]
            col,xa,ya,symbol,line = [curTrace[k] for k in ['color','xaxis','yaxis','symbol','line']]
            # print(col,'-----',xa,'-----',ya,'-----',symbol,'-----',line)
            fig.add_trace(go.Scatter(
                x=df.index,y=df[trace],name=trace,
                mode="lines+markers",xaxis=xa,yaxis=ya,
                marker=dict(color = col,size=10,symbol=symbol),
                line=dict(color = col,dash=line))
            )
        return fig

    def printDFSpecial(self,df,allRows=True):
        # pd.describe_option('col',True)
        colWidthOri = pd.get_option('display.max_colwidth')
        rowNbOri = pd.get_option('display.max_rows')

        pd.set_option('display.max_colwidth',None)
        if allRows :
            pd.set_option('display.max_rows',None)
        pd.set_option('display.max_colwidth',colWidthOri)
        pd.set_option('display.max_rows',rowNbOri)

    def updateStyleGraph(self,fig,style='lines+markers',colmap='jet',heightGraph=700):
        '''
        see self.styles
        '''

        if style=='lines+markers':
            fig.update_traces(mode='lines+markers',line_shape='linear', marker_line_width=0.2, marker_size=6,line=dict(width=3))
        elif style=='markers':
            fig.update_traces(mode='markers', marker_line_width=0.2, marker_size=4)
        elif style=='stairs':
            fig.update_traces(mode='lines+markers',line_shape='hv', marker_line_width=0.2, marker_size=6,line=dict(width=3))
        elif style=='lines':
            fig.update_traces(mode='lines',line=dict(width=1),line_shape='linear')
        self.updateColorMap(fig,colmap)
        fig.update_layout(height=heightGraph)
        return fig

    def quickLayout(self,fig,title='',xlab='',ylab='',style='std'):
        if style=='std':
            fig.update_layout(
                title={'text': title,'x':0.5,'xanchor': 'center','yanchor': 'top'},
                title_font_color="RebeccaPurple",title_font_size=18),

        if style=='latex':
            fig.update_layout(
                font=dict(family="Times New Roman",size=18,color="black"),
                title={'text': title,'x':0.5,'xanchor': 'center','yanchor': 'top'},
                title_font_color="RebeccaPurple",title_font_size=22,
                height=800,width=1200
            )
        if not not xlab:fig.update_layout(yaxis_title = ylab)
        if not not ylab:fig.update_layout(xaxis_title = xlab)
        # fig.show(config=dict({'scrollZoom': True}))
        return fig

    def legendStyle(self,fig,style='big'):
        if style=='plotlySetup':
            return fig.update_layout(
                legend=dict(x=0,y=1,traceorder="reversed",
                font=dict(family="Courier",size=12,color="black"),
                    title_font_family="Times New Roman",
                    bgcolor="LightSteelBlue",bordercolor="Black",borderwidth=2
                    )
                )
        elif style=='big':
            return fig.update_layout(
                font=dict(family="Courier",size=16,color="black"),
                legend=dict(
                            title_font_family="Times New Roman",
                            title_font_size=22,
                            font=dict(family="Courier",size=18,color="black")
                    )
                )
        elif style=='big2':
            fig.update_layout(
                title="Plot Title",
                xaxis_title="X Axis Title",
                yaxis_title="Y Axis Title",
                font=dict(family="Courier New, monospace",size=18,color="black")
    )

    def prepareDFsforComparison(self,dfs,groups,group1='group1',group2='group2',regexpVar='',rs=None,):
        'dfs:list of pivoted dataframes with a timestamp as index'
        dfsOut=[]
        for df,groupy in zip(dfs,groups):
            if not not rs:
                # print(rs)
                df=df.resample(rs).apply(np.mean)
            df['timestamp']= df.index
            df[group1]  = groupy
            df = df.melt(id_vars = ['timestamp',group1])
            if not regexpVar:df[group2] = df.variable
            else:df[group2] = df.variable.apply(lambda x:re.findall(regexpVar,x)[0])
            dfsOut.append(df)
        return pd.concat(dfsOut,axis=0)

    def graphComparaison(self,df,title,ylab,line=True):
        if line : fig=px.line(df,x='timestamp',y='value',color=group2,line_dash=group1)
        else : px.scatter(df,x='timestamp',y='value',color=group2,symbol=group1)
        fig.update_traces(mode='lines+markers',line=dict(width=2))
        fig = self.addTiYXlabs(fig,title='comparaison of ' + title,ylab=ylab,style=1)
        return fig

    def plotGraphType(self,df,typeGraph='scatter',**kwargs):
        if 'tag' in df.columns:
            if typeGraph == 'scatter' :fig = px.scatter(df,x=df.index, y='value', color='tag',**kwargs)
            if typeGraph == 'area' :fig = px.area(df,x=df.index, y='value', color='tag',**kwargs)
            if typeGraph == 'area %' :fig = px.area(df,x=df.index, y='value', color='tag',groupnorm='percent',**kwargs)
        else :
            if typeGraph == 'scatter' :fig = px.scatter(df)
            if typeGraph == 'area' :fig = px.area(df)
            if typeGraph == 'area %' :fig = px.area(df,groupnorm='percent')
        return fig

    def legendPersistant(self,previousFig,newFig):
        if not not previousFig:
            visible_state = {}
            for trace in previousFig['data']:
                idxO = [k for k,v in enumerate(newFig['data']) if v['name']==trace['name']]
                if len(idxO)==1:
                    visible = trace['visible'] if 'visible' in trace.keys() else True
                    newFig['data'][idxO[0]]['visible'] = visible
        return newFig

    def addLogo(self,fig):
        fig.add_layout_image(
            dict(
                source=self.sylfenlogo,
                xref="paper", yref="paper",
                x=0., y=1.02,
                sizex=0.12, sizey=0.12,
                xanchor="left", yanchor="bottom"
            ))
        return fig

    def add_drawShapeToolbar(self,fig):
        fig.update_layout(
            dragmode='drawopenpath',
            newshape_line_color='cyan',
            title_text='Draw a path to separate versicolor and virginica',
            modebar_add=['drawline',
                'drawopenpath',
                'drawclosedpath',
                'drawcircle',
                'drawrect',
                'eraseshape'
               ])

    def showPalettes(self,color):
        colorPalettes = self.colorPalettes
        cols = colorPalettes[color]['hex']
        # cols = cols.loc['Red','Salmon Pink','Cordovan','Tomato','Rosy brown','Scarlet','Bittersweet','Blood red']
        data=[[k,k] for k,c in enumerate(cols)]
        df = pd.DataFrame(data)
        df.index = colorPalettes[color].index
        fig=px.scatter(df.transpose(),color_discrete_sequence=cols)
        fig.update_traces(line_width=20,mode='lines').show()

class DataBase:
    def __init__(self):
        try :
            import psycopg3 as psycopg
        except:
            import psycopg2 as psycopg
        self.psycopg=psycopg

    def connectToPSQLsDataBase(self,connParameters):
        connReq = ''.join([k + "=" + v + " " for k,v in connParameters.items()])
        conn = self.psycopg.connect(connReq)
        return conn

    def connectToDataBase(self,h,p,d,u,w):
        connReq = "host=" + h + " port=" + p + " dbname="+ d +" user="+ u + " password=" + w
        conn    = psycopg.connect(connReq,autocommit=True)
        return conn

    def gettimeSQL(self,secs=10*60):
        t1 = dt.datetime.now()
        # t1 = dt.datetime.now(tz=timezone.utc)
        t0 = t1 - dt.timedelta(seconds=secs)
        timeRange = [t.strftime('%Y-%m-%d %H:%M:%S').replace('T',' ') for t in [t0,t1]]
        return timeRange[0], timeRange[1]

    def readSQLdataBase(self,conn,patSql,secs=60*2,tagCol="tag",tsCol="timestampz"):
        t0,t1 = self.gettimeSQL(secs=secs)
        timeSQL = tsCol + " BETWEEN '" + t0 +"' AND '" + t1 +"'"
        # tagSQL = tagCol + " like '" + patSql + "'"
        tagSQL = tagCol + " ~ '" + patSql + "'"
        sqlQ = "select * from realtimedata where " + timeSQL + " and  " + tagSQL + ";"
        df = pd.read_sql_query(sqlQ,conn,parse_dates=[tsCol])
        return df

    def readSeveralTagsSQL(self,conn,tags,secs=60*2,tagCol="tag",tsCol="timestampz",timeRange=None):
        if not timeRange:
            t0,t1 = self.gettimeSQL(secs=secs)
        else :
            t0,t1 = timeRange
        timeSQL = tsCol + " BETWEEN '" + t0 +"' AND '" + t1 +"'"
        tagSQL = tagCol + " in ('" + "','".join(tags) + "')"
        sqlQ = "select * from realtimedata where " + timeSQL + " and  " + tagSQL + ";"
        df = pd.read_sql_query(sqlQ,conn,parse_dates=[tsCol])
        return df

    def executeSQLRequest(self,conn,sqlR):
        cur  = conn.cursor()
        cur.execute(sqlR)
        data = cur.fetchall()
        for row in data :
            print(row)

    def showAllTables(self,conn):
        sqlR = 'select * from information_schema.tables'
        self.executeSQLRequest(sqlR)

class EmailSmtp:

    import os
    import smtplib
    from email.mime.multipart import MIMEMultipart
    from email.mime.base import MIMEBase
    from email.mime.text import MIMEText
    from email.utils import COMMASPACE
    from email import encoders

    host = None
    port = 25
    user = None
    password = None
    isTls = False

    # Constructor
    def __init__(self, host='127.0.0.1', port=25, user=None, password=None, isTls=False):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.isTls = isTls

    # Send a email with the possibility to attach one or several  files
    def sendMessage(self, fromAddr, toAddrs, subject, content, files=None):
        '''
        Send an email with attachments.

        - Configuring:
            smtp = EmailSmtp()
            smtp.host = 'smtp.office365.com'
            smtp.port = 587
            smtp.user = "datalab@akka.eu"
            smtp.password = "xxxxx"
            smtp.isTls = True

        - Examples of contents:
            # A pure text content
            content1 = "An alert level 3 has been created from the system"
            # Another pure text content
            content2 = [["An alert level 3 has been created from the system", "text"]]
            # A pure html content
            content3 = [["An alert level 3 has been created <br>from the system.<br>", "html"]]
            # A list of text and html contents
            content4 = [
                ["ALERT LEVEL 3!\n", "text"],
                ["An alert level 3 has been created <br>from the system.<br><br>", "html"],
                ["ALERT LEVEL 2!\n", "text"],
                ["An alert level 2 has been also created <br>from the system.<br>", "html"]
            ]

        - Example of attaching file(s):
            # Specifying only one file
            files1 = "./testdata/bank.xlsx"
            # Specifying several files
            files2 = ["./testdata/bank.xlsx", "./testdata/OpenWeather.json"]

        - Example of sending a message:
            # Choose your message and send it
            smtp.sendMessage(
                     fromAddr = "ALERTING <data.intelligence@akka.eu>",
                     toAddrs = ["PhilAtHome <prossblad@gmail.com>", "PhilAtCompany <philippe.rossignol@akka.eu>"],
                     subject = "WARNING: System issue",
                     content = content4,
                     files = files2
            )
        '''

        # Prepare the message
        message = self.MIMEMultipart()
        message["From"] = fromAddr
        message["To"] = self.COMMASPACE.join(toAddrs)
        from email.utils import formatdate
        message["Date"] = formatdate(localtime=True)
        message["Subject"] = subject

        # Create the content (text, html or a combination)
        if (type(content) is not str and type(content) is not list): content = str(content)
        if (type(content) is str): content = [[content, "plain"]]
        for msg in content:
            if (msg[1].strip().lower() != "html"): msg[1] = "plain"
            message.attach(self.MIMEText(msg[0], msg[1]))

        # Attach the files
        if (files != None):
            if (type(files) is str): files = [files]
            for path in files:
                part = self.MIMEBase("application", "octet-stream")
                with open(path, "rb") as file: part.set_payload(file.read())
                self.encoders.encode_base64(part)
                part.add_header("Content-Disposition", 'attachment; filename="{}"'.format(self.os.path.basename(path)))
                message.attach(part)

        # Send the message
        if (fromAddr == None): fromAddr = user
        con = self.smtplib.SMTP(self.host, self.port)
        if (self.isTls): con.starttls()
        if (self.user != None and self.password != None): con.login(self.user, self.password)
        con.sendmail(fromAddr, toAddrs, message.as_string())
        con.quit()

class Callback():
    import base64
    import io

    def __init__(self,):
        self.utils = Utils()

    def parse_contents(self,contents, filename):
        content_type, content_string = contents.split(',')
        decoded = self.base64.b64decode(content_string)
        if 'csv' in filename:
            # Assume that the user uploaded a CSV file
            return pd.read_csv(
                self.io.StringIO(decoded.decode('utf-8')),index_col=0)
        elif 'xls' in filename:
            # Assume that the user uploaded an excel file
            return pd.read_excel(self.io.BytesIO(decoded))

    def df_toTable(self,df):
        rows = df.to_dict('records')
        cols=[{"name": i, "id": i,'renamable': True, 'deletable': True}
                            for i in df.columns]
        return rows,cols

    def exportDataOnClick(self,fig,baseName=None):
        dfs = []
        gofig=go.Figure(fig)
        for trace in gofig.data :
            tmp = pd.DataFrame([trace['x'],trace['y']])
            tmp = tmp.transpose()
            tmp = tmp.set_index(0)
            tmp.columns=[trace.name]
            dfs.append(tmp)
        df = pd.concat(dfs,axis=1)
        filename = 'exportedData'
        return df,filename

class Modebus_utils():
    def quick_modebus_decoder(self,regs,dtype,bo_out='=',bo_in='='):
        if dtype=='INT32' or dtype == 'UINT32':
            value = struct.unpack(bo_out + 'i',struct.pack(bo_in+"2H",*regs))[0]
        if dtype == 'IEEE754':
            value = struct.unpack(bo_out + 'f',struct.pack(bo_in+"2H",*regs))[0]
        elif dtype == 'INT64':
            value = struct.unpack(bo_out + 'q',struct.pack(bo_in+"4H",*regs))[0]
        return value



# #######################
# #      SIMULATORS     #
# #######################
from pymodbus.server.sync import ModbusTcpServer
from pymodbus.datastore import ModbusSequentialDataBlock, ModbusSlaveContext, ModbusServerContext
from pymodbus.client.sync import ModbusTcpClient as ModbusClient
import opcua
import numpy as np
import threading,time

class Simulator():
    ''' for inheritance a simulator should have:
    - inheritance of a Device children class
    - a function "serve" that starts the serveer
    - a function "writeInRegisters" to feed the data.
    - a function "shutdown_server" to shutdown the server.
    '''
    def __init__(self,speedflowdata=500,volatilitySimu=5):
        '''
        - speedflowdata : single data trigger event in ms
        - volatilitySimu : how much random variation (absolute units)
        '''
        self.volatilitySimu = volatilitySimu
        self.speedflowdata = speedflowdata
        # self.server_thread = threading.Thread(target=self.serve)
        # self.server_thread.daemon = True
        self.feed = True
        self.stopfeed = threading.Event()
        self.feedingThread = threading.Thread(target=self.feedingLoop)
        # self.feedingThread.daemon = True

    def stop(self):
        self.stopfeed.set()
        self.shutdown_server()
        print("Server stopped")

    def start(self):
        print("Start server...")
        # self.server_thread.start()
        self.serve()
        print("Server is online")
        # self.feedingLoop()
        # self.feedingThread.start()
        # print("Server simulator is feeding")

    def stopFeeding(self):
        self.feed=False

    def startFeeding(self):
        self.feed=True

    def feedingLoop(self):
        while not self.stopfeed.isSet():
            if self.feed:
                start=time.time()
                self.writeInRegisters()
                print('fed in {:.2f} milliseconds'.format((time.time()-start)*1000))
                time.sleep(self.speedflowdata/1000 + np.random.randint(0,1)/1000)

    def is_serving(self):
        return self.server_thread.is_alive()

class SimulatorModeBus(Simulator):
    ''' can only be used with a children class inheritating from a class that has
    attributes and methods of Device.
    ex : class StreamVisuSpecial(ComConfigSpecial,SimulatorModeBus)
    with class ComConfigSpecial(Device)'''

    def __init__(self,bo='=',*args,**kwargs):
        '''
        - bo : byteorder : bigendian >, littleendian <, native =, network(big endian) !
        '''
        self.bo = bo
        Simulator.__init__(self,*args,**kwargs)
        # initialize server with random values
        self.dfInstr['value']=self.dfInstr['type'].apply(lambda x:np.random.randint(0,100000))
        self.dfInstr['precision']=0.1
        self.dfInstr['FREQUENCE_ECHANTILLONNAGE']=1
        allTCPid = list(self.dfInstr['addrTCP'].unique())
        # Create an instance of ModbusServer
        slaves={}
        for k in allTCPid:
            slaves[k]  = ModbusSlaveContext(hr=ModbusSequentialDataBlock(0, [k]*128))
            self.context = ModbusServerContext(slaves=slaves, single=False)
        self.server = ModbusTcpServer(self.context, address=("", self.port))
        self.server_thread = threading.Thread(target=self.serve)
        self.server_thread.daemon = True

    def start(self):
        print("Start server...")
        self.server_thread.start()
        print("Server is online")
        self.feedingThread.start()
        print("Server simulator is feeding")

    def generateRandomData(self,idTCP):
        ptComptage = self.dfInstr[self.dfInstr['addrTCP']==idTCP]
        byteflow=[]
        values=[]
        # te = ptComptage.iloc[[0]]
        for tag in ptComptage.index:
            te = ptComptage.loc[tag]
            # print(te)
            # address = te.index[0]
            typevar = te.type
            if typevar=='INT32':
                value = int(te.value + np.random.randint(0,value*self.volatilitySimu/100))
                # conversion of long(=INT32) into 2 shorts(=DWORD=word)
                valueShorts = struct.unpack(self.bo + '2H',struct.pack(self.bo + "i",value))
            if typevar=='INT64':
                value = int(te.value + np.random.randint(0,value*self.volatilitySimu/100))
                try:
                    # conversion of long long(=INT64) to 4 short(=DWORD=word)
                    valueShorts = struct.unpack(self.bo + '4H', struct.pack(self.bo + 'q',value))
                except:
                    print(value)
            if typevar=='IEEE754':
                value = te.value + np.random.randn()*te.value*self.volatilitySimu/100
                # value = 16.565
                # conversion of float(=IEEE7554O) to 2 shorts(=DWORD)
                valueShorts=struct.unpack(self.bo + '2H', struct.pack(self.bo+'f', value))
            byteflow.append(valueShorts)
            self.dfInstr.loc[tag,'value']=value
            # values.append(value)
        # return [l for k in byteflow for l in k],values
        return [l for k in byteflow for l in k]

    def writeInRegisters(self):
        feedingClient = ModbusClient(host='localhost',port=self.port)
        feedingClient.connect()

        for idTCP in self.allTCPid:
            # print(idTCP)
            ptComptage = self.dfInstr[self.dfInstr['addrTCP']==idTCP]

            # #######
            #                   IMPORTANT CHECK HERE
            #           block should have continuous adresses with no gap!
            # #######
            # ptComptage = ptComptage[ptComptage.intAddress<10000]
            try:
                byteflow = self.generateRandomData(idTCP)
                feedingClient.write_registers(ptComptage.intAddress[0],byteflow,unit=idTCP)
            except:
                print(dt.datetime.now().astimezone())
                print('***********************************')
                print(str(idTCP) + 'problem generating random Data')
                traceback.print_exc()
                print('***********************************')

        tagtest='C00000001-A003-1-kW sys-JTW'
        print(tagtest + ' : ',self.dfInstr.loc[tagtest,'value'])
        feedingClient.close()

    def serve(self):
        self.server.serve_forever()

    def shutdown_server(self):
        self.server.shutdown()
        print("Server simulator is shutdown")

class SimulatorOPCUA(Simulator):
    '''
    dfPLC should have columns index as tags, DATATYPE
    '''
    def __init__(self,endpointUrl,dfPLC,nameSpace,*args,**kwargs):
        start=time.time()
        Simulator.__init__(self,*args,**kwargs)
        time.time()-start
        self.server=opcua.Server()
        self.endpointUrl=endpointUrl
        self.nameSpace=nameSpace
        self.server.set_endpoint(endpointUrl)
        self.server.register_namespace('room1')
        self.dfPLC=dfPLC
        self.dfPLC.MIN=dfPLC.MIN.fillna(-20000)
        self.dfPLC.MAX=dfPLC.MAX.fillna(20000)
        self.nodeValues = {}
        self.nodeVariables = {}
        self.createNodes()

    def serve(self):
        try:
            print("start server")
            self.server.start()
            print("server Online")
        finally:
            self.shutdown_server()
    def shutdown_server(self):
        self.server.stop()
        print("server Offline")

    def createRandomInitalTagValues(self):
        valueInit={}
        for tag in list(self.dfPLC.index.unique()):
            tagvar=self.dfPLC.loc[tag]
            if tagvar.DATATYPE=='STRING(40)':
                valueInit[tag] = 'STRINGTEST'
            else:
                valueInit[tag] = np.random.randint(tagvar.MIN,tagvar.MAX)
        return valueInit

    def createNodes(self):
        objects =self.server.get_objects_node()
        beckhoff=objects.add_object(self.nameSpace,'beckhof')
        valueInits = self.createRandomInitalTagValues()
        for tag,val in valueInits.items():
            self.nodeValues[tag]    = val
            self.nodeVariables[tag] = beckhoff.add_variable(self.nameSpace+tag,tag,val)

    def writeInRegisters(self):
        for tag,var in self.nodeVariables.items():
            tagvar=self.dfPLC.loc[tag]
            if tagvar.DATATYPE=='REAL':
                newValue = self.nodeValues[tag] + self.volatilitySimu*np.random.randn()*tagvar.PRECISION
                self.nodeVariables[tag].set_value(newValue)
            else:
                newValue = np.random.randint(0,2)
                self.nodeVariables[tag].set_value(newValue)
            self.nodeValues[tag] = newValue
        # tagTest = 'SEH1.STB_STK_03.HER_01_CL01.In.HR26'
        # tagTest = 'SEH1.GWPBH_PMP_05.HO00'
        tagTest = 'SEH1.STB_GFC_00_PT_01_HC21'
        # tagTest = 'SEH1.STB_STK_01.SN'
        # tagTest = 'SEH1.HPB_STG_01a_HER_03_JT_01.JTVAR_HC20'
        print(tagTest + ': ',self.nodeVariables[tagTest].get_value())



# #######################
# #  VERISONS MANAGER    #
# #######################
import collections
from . import comUtils
from .comUtils import (print_file)
from .utilsD import Utils
import glob,re,pandas as pd,os,time,pickle
sort_list=lambda x:list(pd.Series(x).sort_values())
import plotly.express as px
import plotly.graph_objects as go
import subprocess as sp

class VersionsManager():
    def __init__(self,folderPkl,dir_plc,file_transition,pattern_plcFiles='*plc*.csv',ds=None):
        '''
        plc versions filenames should be of the form x.y or x with the letter v or version in front of it.
        '''
        self._streamer     = comUtils.Streamer()
        self._fs           = comUtils.FileSystem()
        self.dir_plc       = dir_plc
        self._versionFiles = glob.glob(self.dir_plc+pattern_plcFiles)
        self.versions      = {re.search(r'\d+(\.\d+)?',f.split('/')[-1]).group():f for f in self._versionFiles}
        self.folderPkl     = folderPkl
        self.file_transitions = file_transition
        self.transitions    = pd.ExcelFile(self.file_transitions).sheet_names
        self._file_versionning = self.dir_plc+'versions_data.pkl'
        self.versions = pd.DataFrame([{'file':f,'version':v} for v,f in self.versions.items()]).set_index('version').sort_index()
        self.versions.index=self.versions.index.astype(float)
        self._ds=ds
        if not os.path.exists(self._file_versionning):self.generate_versionning_data()
        [self.df_plcs,self.df_transition_rules] = Utils().loads_pickle(self._file_versionning)

    ##########################
    # private init functions #
    ##########################
    def _load_transition_rules(self):
        rules={}
        for t in self.transitions:
            tmp=pd.read_excel(self.file_transitions,sheet_name=t)
            idx_nana = tmp[tmp.isna().all(axis=1)].index.min()
            rules[t]=tmp.iloc[:idx_nana,:]
        return rules

    def _load_PLC_versions(self):
        print_file('Start reading all .xlsm files....')
        df_plcs = {}
        for v,f in self.versions['file'].to_dict().items():
            print_file(f)
            df_plcs[v] = pd.read_excel(f,index_col=0)
        print_file('')
        print_file('concatenate tags of all dfplc verion')
        return df_plcs

    def _load_missing_tags_map(self,d0=None,d1=None,v0=None,v1=None):
        list_days=self._fs.listfiles_folder(self.folderPkl)
        list_days.sort()

        if d0 is None:d0=list_days[0]
        if d1 is None:d1=list_days[-1]
        vi=self.versions.index
        if v0 is None:v0=vi[0]
        if v1 is None:v1=vi[-1]

        versions=[k for k in vi[(vi>=float(v0))&(vi<=float(v1))]]
        list_days=[d for d in list_days if d in pd.date_range(d0,d1)]

        df={}
        for d in list_days:
            tmp=self.missing_tags_versions(d,versions)
            df[d]=pd.Series({k:len(v) for k,v in tmp.items()})
        return pd.concat(df,axis=1).T


    #####################
    # public methods    #
    #####################
    def generate_versionning_data(self,plcs=True,transition_rules=True):
        print('generating versionning data and store in :'+self._file_versionning)
        if os.path.exists(self._file_versionning):[dfplcs,df_transition_rules] = Utils().loads_pickle(self._file_versionning)

        start=time.time()

        if plcs:dfplcs=self._load_PLC_versions()

        if transition_rules:transition_rules=self._load_transition_rules()

        f = open(self._file_versionning,'wb')
        pickle.dump(dfplcs,f)
        pickle.dump(transition_rules,f)
        # self.df_nbTagsFolder = loadconfFile(self.file_df_nbTags,self.load_nbTags_folders,buildFiles[1])
        # self.map_missingTags,self.map_missingTags_len = loadconfFile(self.file_map_missingTags,self.load_missingTags_versions,buildFiles[2])
        f.close()
        [self.df_plcs,self.df_transition_rules] = Utils().loads_pickle(self._file_versionning)
        print('finish loading versionning data \n'+'='*60+'\n')

    def get_rename_tags_map_from_rules(self,transition,show_diagnostique=True):
        def pattern_to_regexp(s,back_ref=False):
            if back_ref:
                return s.apply(lambda x:x.replace('xxx','\\1'))
            else:
                return s.apply(lambda x:x.replace('xxx','(.*)'))

        vold,vnew=[float(k) for k in transition.split('_')]
        df_rules_transition=self.df_transition_rules[transition]

        patterns_new = pattern_to_regexp(df_rules_transition[df_rules_transition['old_pattern'].isna()]['new_pattern']).to_list()
        patterns_old = pattern_to_regexp(df_rules_transition[df_rules_transition['new_pattern'].isna()]['old_pattern']).to_list()


        df_compare=self.compare_plcs(vold,vnew)
        if df_compare.empty:
            print('='*60,'\nLes 2 versions',vold,'et', vnew,'ont exactement les même tags.\n Rien à faire mon ami.')
            sys.exit()
        r=df_compare[[k for k in df_compare.columns if 'removed' in k]].dropna().squeeze(1)
        a=df_compare[[k for k in df_compare.columns if 'added' in k]].dropna().squeeze(1)

        ################################
        #  find tags to add and remove #
        #       from  patterns         #
        ################################
        if len(patterns_new)>0:
            tags_added=pd.concat([a[a.str.contains(pat)] for pat in patterns_new])
        else:
            tags_added=pd.Series([],name=a.name)
        if len(patterns_old)>0:
            tags_removed=pd.concat([r[r.str.contains(pat)] for pat in patterns_old])
        else:
            tags_removed=pd.Series([],name=r.name)

        ##### find tags to rename from patterns
        a2 = a[~a.isin(tags_added)]
        r2 = r[~r.isin(tags_removed)]
        dd=df_rules_transition.dropna()
        patternMap_rename = pd.concat([pattern_to_regexp(dd['old_pattern']),pattern_to_regexp(dd['new_pattern'],True)],axis=1)


        dfs={}
        for x in range(len(patternMap_rename)):
            pat_old,pat_new=patternMap_rename.iloc[x]
            # print(pat_old,'+',pat_new)
            tags2rename=r2[r2.str.contains(pat_old)].to_list()
            dfs[pat_old]=pd.DataFrame({'old_names':tags2rename,'new_names_from_rules':[re.sub(pat_old,pat_new,t) for t in tags2rename]})

        if len(dfs)>0:
            tags_renamed_map=pd.concat(dfs.values())
        else:
            tags_renamed_map=pd.DataFrame({'old_names':[],'new_names_from_rules':[]})


        t1=tags_added.to_frame()
        t1.columns=['new_names_from_rules']
        t1['old_names']=''
        t2=tags_removed.to_frame()
        t2.columns=['old_names']
        t2['new_names_from_rules']=''
        tags_renamed_map=pd.concat([tags_renamed_map,t1,t2])
        tags_renamed_map['in_new_plc']=tags_renamed_map['new_names_from_rules'].isin(a)

        ##### still to explain
        a3 = pd.Series([t for t in a2 if t not in tags_renamed_map['new_names_from_rules'].to_list()],name='still in '+str(vnew))
        r3 = pd.Series([t for t in r2 if t not in tags_renamed_map['old_names'].to_list()],name='still in '+str(vold))
        still2explain=pd.concat([r3,a3],axis=1)


        ######################
        ## show diagnostique #
        ######################
        if show_diagnostique:
            def std_div(h,t,color=None):
                return '<div style="margin:10px;background-color:None">'+'<h2 style="text-align:center">'+h+'</h2>' + t + '</div>\n'

            t1=df_compare.to_html()
            h1='comparaison des fichiers PLCS'
            t2=tags_renamed_map.to_html()
            h2='table de renommage à partir des règles'
            t3=still2explain.to_html()
            h3='tags restants non trouvés par les règles de renommage'

            f=open('/tmp/table.html','w')
            text='<div style="display:flex">\n'+std_div(h1,t1)+std_div(h2,t2)+std_div(h3,t3)+'</div>'
            f.write(text)
            f.close()
            sp.run('firefox /tmp/table.html',shell=True)

        return tags_renamed_map
        sys.exit()

    def compare_plcs(self,vold,vnew):
        plc_old  = self.df_plcs[vold]
        plc_new  = self.df_plcs[vnew]
        if self._ds is None:
            oldtags = list(plc_old.index)
            newtags = list(plc_new.index)
        else:
            oldtags = list(plc_old[plc_old.DATASCIENTISM==self._ds].index)
            newtags = list(plc_new[plc_new.DATASCIENTISM==self._ds].index)

        print(plc_new)
        r = pd.Series([t for t in oldtags if t not in newtags],name='removed_from_' + str(vold))
        a = pd.Series([t for t in newtags if t not in oldtags],name='added_in_' + str(vnew))
        return pd.concat([r,a],axis=1)

    def _find_presence_tags(self,tags,empty_df=True):
        '''
        empty_df : [boolean] if true will return -1 if not in 0 if empty or 1 if in
        '''
        listDays=os.listdir(self.folderPkl)
        df=pd.DataFrame()
        if empty_df:
            for t in tags:
                vals={}
                for d in listDays:
                    vals[d]=-1
                    if t+'.pkl' in os.listdir(self.folderPkl+d):
                        vals[d]=pd.read_pickle(self.folderPkl+d + '/' + t + '.pkl').empty
                df[t]=vals.values()
        else:
            for t in tags:
                df[t] = [True if t+'.pkl' in os.listdir(self.folderPkl+d) else False for d in listDays]
        df=df.mask(df==False,1).mask(df==True,0)
        df.index=listDays
        df=df.sort_index().astype(int)
        return df

    def presence_tags(self,*args,**kwargs):
        df=self._find_presence_tags(*args,**kwargs)

        fig=px.line(df);fig.update_traces(line_shape='hv',mode='lines+markers')
        for k,t in enumerate(fig.data):t['y']=t['y']+0.01*float((k+1)//2)*((-1)**k)

        return fig

    def presence_tags_transition(self,transition,**kwargs):
        tags_renamed_map=self.get_rename_tags_map_from_rules(transition,False)
        df_rename_tags=tags_renamed_map.iloc[:,:-1]
        df_rename_tags.columns=['old_tag','new_tag']

        df_tags_new = self._find_presence_tags(df_rename_tags[(df_rename_tags['old_tag']=='')]['new_tag'].to_list(),**kwargs).melt(ignore_index=False)
        df_tags_new['group'] = 'tag v2(added)'
        df_tags_removed = self._find_presence_tags(df_rename_tags[(df_rename_tags['new_tag']=='')]['old_tag'].to_list(),**kwargs).melt(ignore_index=False)
        df_tags_removed['group']       = 'tag v1(deleted)'
        df_tags_v2_from_v1    = df_rename_tags[df_rename_tags.all(axis=1)]
        df_v2_renamed_from_v1 = self._find_presence_tags(df_tags_v2_from_v1['new_tag'],**kwargs).melt(ignore_index=False)
        df_v2_renamed_from_v1['group'] = 'tag v2 renamed from v1'
        df_v1_renamed_to_v2   = self._find_presence_tags(df_tags_v2_from_v1['old_tag'],**kwargs).melt(ignore_index=False)
        df_v1_renamed_to_v2['group']   = 'tag v1 renamed to v2'

        df = pd.concat([df_tags_new,df_tags_removed,df_v2_renamed_from_v1,df_v1_renamed_to_v2])
        ######### make graph
        col_dis_map={
        'tag v2(added)':'blue',
        'tag v1(deleted)':'yellow',
        'tag v2 renamed from v1':'green',
        'tag v1 renamed to v2':'red'
        }
        fig=px.line(df,y='value',color='group',symbol='variable',color_discrete_map=col_dis_map);
        fig.update_traces(line_shape='hv',mode='lines+markers')
        v_old,v_new=[float(k) for k in transition.split('_')]
        v1_start=self.versions.loc[v_old,'start']
        v2_start=self.versions.loc[v_new,'start']
        fig.add_vline(x=v1_start)
        fig.add_vline(x=v2_start)

        add_v_number= lambda x,y,t,ax,fig: fig.add_annotation(
                x=x,y=y,text='start of v' + str(t),
                font=dict(family="Courier New, monospace",size=16,color="#000000"),
                align="center",arrowhead=2,arrowsize=1,arrowwidth=2,arrowcolor="#000000",showarrow=True,
                bordercolor="#c7c7c7",borderwidth=2,borderpad=4,bgcolor="#ffffff",opacity=0.8,
                ax=ax,ay=130,
                )

        fig=add_v_number(v1_start,0,v_old,-60,fig)
        fig=add_v_number(v2_start,0.25,v_new,-60,fig)
        try:
            v_next=self.versions.index[[k for k,v in enumerate(self.versions.index) if v==v_new][0]+1]
            v3_start=self.versions.loc[v_next,'start']
            fig.add_vline(x=v3_start)
            fig.add_vrect(x0=v2_start, x1=v3_start, line_width=0, fillcolor="red", opacity=0.2)
            fig=add_v_number(v3_start,0.5,v_next,60,fig)
        except:
            print('v_new :v ',v_new,'is the last version available.')
        ### separate a bit the traces
        for k,t in enumerate(fig.data):t['y']=t['y']+0.01*float((k+1)//2)*((-1)**k)
        return fig

    def get_lines(self):
        lines=[]
        for v in self.versions.index:
            s=pd.Series(self.df_plcs[v].index)
            tmp=(s[s.str.contains('L\d+')]).apply(lambda x:re.findall('L\d+',x)[0])
            lines+=[pd.Series(tmp.unique(),name=v).sort_values().reset_index()[v]]

        l=pd.concat(lines,axis=1)
        all_lines=pd.Series(pd.concat([l[v] for v in l.columns]).unique()).sort_values().dropna()
        l.index=all_lines
        # return l
        d=pd.DataFrame()
        for v in l.columns:
            # d[v]=all_lines.apply(lambda x:x if x in l[v].to_list() else np.nan)
            d[v]=all_lines.apply(lambda x:1 if x in l[v].to_list() else 0)
        d.index=l.index
        d=d.T
        d=d.iloc[[len(d)-k-1 for k in range(len(d))],:]
        return d

    def missing_tags_versions(self,day,versions=None):
        folder=self.folderPkl+day+'/'
        print(folder)
        listTags = [k.split('.pkl')[0] for k in self._fs.listfiles_folder(folder)]
        dfs, tagNotInVersion, tagNotInFolder={},{},{}
        dayCompatibleVersions = {}

        if versions is None:versions=self.df_plcs.keys()
        if isinstance(versions,str) or isinstance(versions,float):versions=[versions]
        for version,dfplc in {v:df for v,df in self.df_plcs.items() if v in versions}.items():
            # keep only valid tags
            if self._ds is None:
                tagsVersion = list(dfplc.index)
            else:
                tagsVersion = list(dfplc.index[dfplc.DATASCIENTISM==self._ds])

            tagNotInVersion[version] = [k for k in listTags if k not in tagsVersion]
            tagNotInFolder[version] = [k for k in tagsVersion if k not in listTags]
            dfs[version] = tagsVersion
            dayCompatibleVersions[version] = tagNotInFolder[version]
        return dayCompatibleVersions

    def show_map_of_compatibility(self,missing_tags_map,binaire=False,zmax=None):
        if zmax is None:
            zmax = missing_tags_map.max().max()
        reverse_scale=True
        # missing_tags_map=missing_tags_map.applymap(lambda x:np.random.randint(0,zmax))
        if binaire:
            missing_tags_map=missing_tags_map.applymap(lambda x:1 if x==0 else 0)
            zmax=1
            reverse_scale=False

        fig=go.Figure(go.Heatmap(z=missing_tags_map,x=['v' + str(k) for k in missing_tags_map.columns],
            y=missing_tags_map.index,colorscale='RdYlGn',reversescale=reverse_scale,
            zmin=0,zmax=zmax))
        fig.update_xaxes(side="top",tickfont_size=35)
        fig.update_layout(font_color="blue",font_size=15)
        fig.show()
        return fig

    ######################
    # make it compatible #
    ######################
    def _create_newtags_folder(self,day,tags):
        folder=self.folderPkl+day+'/'
        if not os.path.exists(folder):os.mkdir(folder)
        df = pd.Series(name='value')
        for tag in tags:
            tagpkl=folder + tag + '.pkl'
            if not os.path.exists(tagpkl):
                df.to_pickle(tagpkl)

    def _replace_tags_folder(self,day,tags2replace):
        folder=self.folderPkl+day+'/'
        for oldtag,newtag in zip(tags2replace['old_tag'],tags2replace['new_tag']):
            oldpath=folder + oldtag+'.pkl'
            newpath=folder + newtag+'.pkl'
            if not os.path.exists(newpath):
                os.rename(oldpath,newpath)
            else:
                s=pd.read_pickle(newpath)
                if s.empty and os.path.exists(oldpath):
                    os.rename(oldpath,newpath)

    def _apply_map_rename_tags(self,d0,d1,transition):
        #### get map of tags that should be renamed
        tags_renamed_map=self.get_rename_tags_map_from_rules(transition)
        df_rename_tags=tags_renamed_map.iloc[:,:-1]
        df_rename_tags.columns=['old_tag','new_tag']
        df_replace_tags=df_rename_tags[df_rename_tags.all(axis=1)]
        #### get map of tags that should be added
        tags_to_add=df_rename_tags[df_rename_tags['old_tag']=='']['new_tag'].to_list()
        for d in pd.date_range(d0,d1):
            day=d.strftime('%Y-%m-%d')
            self._replace_tags_folder(day,df_replace_tags)
            self._create_newtags_folder(day,tags_to_add)

    def make_period_compatible_from_transition(self,d0,d1,transition,execute=True):
        range_days=[(pd.Timestamp(t)+pd.Timedelta(days=d)).strftime('%Y-%m-%d') for t,d in zip([d0,d1],[-2,2])]
        v_old,vnew=[float(k) for k in transition.split('_')]
        idx_v = [k for k,v in enumerate(self.versions.index) if v==v_old][0]
        idx_v0=max(0,idx_v-1)
        v0=self.versions.index[idx_v0]

        fig=self.presence_tags_transition(transition)
        fig.update_xaxes(range=range_days).show()
        fig.update_layout(title_text='before remapping')
        df_missing_tags=self._load_missing_tags_map(d0,d1,v0=v0)

        self.show_map_of_compatibility(df_missing_tags)

        if not execute:return
        #########################
        # execute the remapping #
        #########################
        self._apply_map_rename_tags(d0,d1,transition)
        #### check that the transformation is correct and there are no tags missing

        fig=self.presence_tags_transition(transition)
        fig.update_layout(title_text='after remapping')
        fig.update_xaxes(range=range_days).show()

        df_missing_tags=self._load_missing_tags_map(d0,d1,v0=v0)
        self.show_map_of_compatibility(df_missing_tags)

    def create_missing_tags_period_version(self,d0,d1,version):
        dfplc=self.df_plcs[version]
        if self._ds is None:
            list_tags=dfplc.index.to_list()
        else:
            list_tags=dfplc[dfplc.DATASCIENTISM==self._ds].index.to_list()
        for d in pd.date_range(d0,d1):
            day=d.strftime('%Y-%m-%d')
            # print(day)
            self._create_newtags_folder(day,list_tags)

    #######################
    #       UTILS         #
    #######################
    def _totime(self,x):
        y=x.split('/')
        return pd.Timestamp(y[0]+' ' +y[1] + ':' + y[2])

    def are_tags_in_PLCs(self,tags,ds=True):
        df_plcs = self.df_plcs
        if not not ds:
            df_plcs = {k:v[v.DATASCIENTISM==ds] for k,v in self.df_plcs.items()}
        tagInplc={}
        for tag in tags:
            tagInplc[tag]=[True if tag in list(v.index) else False for k,v in df_plcs.items()]
        return pd.DataFrame.from_dict(tagInplc,orient='index',columns=df_plcs.keys()).T.sort_index()

    def patterntags_inPLCs(self,pattern,ds=True):
        df_plcs = self.df_plcs
        if not not ds:
            df_plcs = {k:v[v.DATASCIENTISM==ds] for k,v in self.df_plcs.items()}
        patterntags_plcs={}
        # print(df_plcs.keys())
        for v,dfplc in df_plcs.items():
            # return pd.DataFrame.from_dict(tagInplc,orient='index',columns=df_plcs.keys()).T.sort_index()
            patterntags_plcs[v]=list(dfplc.index[dfplc.index.str.contains(pattern)])
            patterntags_plcs = collections.OrderedDict(sorted(patterntags_plcs.items()))
        return patterntags_plcs

class VersionsManager_minutely(VersionsManager):
    #######################
    # GENERATE DATAFRAMES #
    #######################
    def load_nbTags_folders(self):
        # get_lentags=lambda x:len(self._fs.listfiles_folder(x))
        df_nbtags=self._streamer.actionMinutes_pooled(self.tmin,self.tmax,self.folderData,self.get_lentags)
        return pd.DataFrame.from_dict(df_nbtags,orient='index')

    def load_missingTags_versions(self,period=None,pool=True):
        '''-period : [tmin,tmax] timestamps'''
        map_missingTags = self._compute_all_minutefolders(self.get_missing_tags_versions,period=period)
        map_missingTags = pd.DataFrame(map_missingTags).T
        map_missingTags_len = map_missingTags.applymap(lambda x:len(x))
        return map_missingTags,map_missingTags_len

    def load_presenceTags(self,period=None,frequence='daily'):
        if frequence=='daily':
            return self._compute_all_minutefolders(self.get_presenceTags_folder,period=period)
        elif frequence=='minutely':
            return self._compute_all_minutefolders(self.get_presenceTags_folder,period=period)

    def _compute_all_minutefolders(self,function,*args,period=None,pool=True):
        '''-period : [tmin,tmax] timestamps'''
        if period is None:
            tmin = self.tmin
            tmax = self.tmax
        else :
            tmin,tmax=period
        if tmax - tmin -dt.timedelta(days=2)<pd.Timedelta(seconds=0):
            pool=False
        df = self._streamer.actionMinutes_pooled(tmin,tmax,self.folderData,function,*args,pool=pool)
        # print(df)
        df = pd.DataFrame(df).T
        df.index=[self.totime(x) for x in df.index]
        return df

    def make_it_compatible_with_renameMap(self,map_renametag,period):
        ## from one version to an adjacent version (next or last):
        ## get transition rules
        # self.get_transition_rename_rules(transition)
        ## get the corresponding map of tags that should be renamed
        ## rename tags that should be renamed
        '''map_renametag :
            - should be a dataframe with columns ['oldtag','newtag']
            - should have None values in oldtag column for brand newtags
            - should have None values in newtag column for deleted tags
            '''
        tags2replace = map_renametag[map_renametag.apply(lambda x:not x['oldtag']==x['newtag'] and not x['oldtag'] is None and not x['newtag'] is None,axis=1)]
        print()
        print('MAP OF TAGS TO REPLACE '.rjust(75))
        print(tags2replace)
        print()
        replacedmap=self._compute_all_minutefolders(self.get_replace_tags_folder,tags2replace,period=period)
        print()
        print('MAP OF REPLACED TAGS '.rjust(75))
        print(replacedmap)
        return replacedmap

class VersionsManager_daily(VersionsManager):
    def __init__(self,*args,**kwargs):
        VersionsManager.__init__(self,*args,**kwargs)
        # self._streamer.actiondays(self.tmin,self.tmax,self.folderData,self._streamer.create_dayfolder,pool=False)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
dirParent = "dorianUtils"

setuptools.setup(
name="dorianUtils", # Replace with your own username
version="6.3.1",
author="Dorian Drevon",
author_email="drevondorian@gmail.com",
description="Utilities package",
long_description=long_description,
long_description_content_type="text/markdown",
classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
],
packages=setuptools.find_packages(),
package_data={'': ['conf/*']},
include_package_data=True,
install_requires=['IPython==7.20.0','pandas==1.3.1','psycopg2-binary','odfpy==1.4.1','plotly>=5.5.0',
    'pymodbus==2.5.3','opcua==0.98.13','cryptography==2.8','Pillow==7.0.0','openpyxl==3.0.7','psutil==5.8.0',
	'colorama==0.4.3']
,python_requires=">=3.8"
)
